# !/usr/bin/env python3
# -*- coding: utf-8 -*-

import os
import tempfile

from neurostatx.io.download import get_home, download_data, save_files_dict


download_data(save_files_dict(), keys=["data.zip"])
tmp_dir = tempfile.TemporaryDirectory()


def test_help(script_runner):
    ret = script_runner.run(["AddNodesAttributes", "-h"])

    assert ret.success


def test_set_nodes_attributes(script_runner):
    os.chdir(os.path.expanduser(tmp_dir.name))
    in_graph = os.path.join(get_home(), "data/graph_file.gml")
    in_df = os.path.join(get_home(), "data/clusters_membership_3.xlsx")
    out_file = os.path.join(get_home(), "data/graph_with_attributes.gml")

    ret = script_runner.run([
        "AddNodesAttributes",
        "--in-graph", in_graph,
        "--in-dataset", in_df,
        "--labels", "gestage",
        "--labels", "age",
        "--labels", "iq",
        "--id-column", "subjectkey",
        "--out-file", out_file,
        "-f", "-v"]
    )

    assert ret.success



import unittest
import networkx as nx
import numpy as np
import matplotlib.pyplot as plt

from neurostatx.network.viz import (creating_node_colormap,
                                    create_cmap_from_list,
                                    set_nodes_position)


class TestNodeColormapFunctions(unittest.TestCase):

    def test_creating_node_colormap(self):
        percentile_dict = {'cluster1': [0, 1, 0, 1, 0],
                           'cluster2': [0, 0, 1, 0, 0]}
        expected_output = ['black', np.array(plt.cm.tab10(1)),
                           np.array(plt.cm.tab10(1)),
                           np.array(plt.cm.tab10(1)), 'black']
        for i in range(0, len(expected_output)):
            if isinstance(expected_output[i], np.ndarray):
                self.assertEqual(
                    expected_output[i].all(),
                    creating_node_colormap(percentile_dict)[i].all())
            else:
                self.assertEqual(expected_output[i],
                                 creating_node_colormap(percentile_dict)[i])

    def test_create_cmap_from_list_integer_array(self):
        integer_array = np.array([0, 1, 2, 3, 4])
        expected_output = ['darkgrey',
                           np.array(plt.cm.tab10(1)),
                           np.array(plt.cm.tab10(2)),
                           np.array(plt.cm.tab10(3)),
                           np.array(plt.cm.tab10(4))]
        for i in range(0, len(expected_output)):
            if isinstance(expected_output[i], np.ndarray):
                self.assertEqual(
                    expected_output[i].all(),
                    create_cmap_from_list(integer_array)[i].all())
            else:
                self.assertEqual(expected_output[i],
                                 create_cmap_from_list(integer_array)[i])

    def test_create_cmap_from_list_float_array(self):
        float_array = np.array([0.1, 0.2, 0.3, 0.4, 0.5])
        expected_output = [np.array(plt.cm.plasma(0)),
                           np.array(plt.cm.plasma(1)),
                           np.array(plt.cm.plasma(2)),
                           np.array(plt.cm.plasma(3)),
                           np.array(plt.cm.plasma(4))]
        for i in range(0, len(expected_output)):
            if isinstance(expected_output[i], np.ndarray):
                self.assertEqual(
                    expected_output[i].all(),
                    create_cmap_from_list(float_array)[i].all())
            else:
                self.assertEqual(expected_output[i],
                                 create_cmap_from_list(float_array)[i])


class TestSetNodesPosition(unittest.TestCase):
    def test_set_nodes_position(self):
        # Create a graph
        G = nx.Graph()
        G.add_nodes_from([1, 2, 3])

        # Define positions
        pos = {1: (0, 0), 2: (1, 1), 3: (2, 2)}

        # Call the function
        set_nodes_position(G, pos)

        # Check if node positions are set correctly
        self.assertTrue(all("pos" in G.nodes[node] for node in G.nodes))
        for node, position in pos.items():
            self.assertAlmostEqual(G.nodes[node]["pos"][0],
                                   float(position[0]))
            self.assertAlmostEqual(G.nodes[node]["pos"][1],
                                   float(position[1]))


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import coloredlogs
import logging
import os
import warnings

from cyclopts import App, Parameter
import matplotlib.pyplot as plt
import networkx as nx
import numpy as np
import pandas as pd
from sklearn.preprocessing import scale
from typing import List
from typing_extensions import Annotated

from neurostatx.io.utils import (assert_input, assert_output_dir_exist)
from neurostatx.network.utils import fetch_attributes_df, fetch_edge_data
from neurostatx.statistics.models import (plsr_cv, permutation_testing,
                                          ScoringMethod)
from neurostatx.io.viz import generate_coef_plot, flexible_hist

# Initializing the app.
app = App(default_parameter=Parameter(negative=()))


@app.default()
def PartialLeastSquareRegression(
    in_graph: Annotated[
        str,
        Parameter(
            show_default=False,
            group="Essential Files Options",
        ),
    ],
    out_folder: Annotated[
        str,
        Parameter(
            show_default=False,
            group="Essential Files Options",
        ),
    ],
    attributes: Annotated[
        List[str],
        Parameter(
            group="Model Parameters",
            show_default=True,
        ),
    ] = None,
    weight: Annotated[
        str,
        Parameter(
            group="Model Parameters",
            show_default=True,
        ),
    ] = "membership",
    splits: Annotated[
        int,
        Parameter(
            group="Model Parameters",
            show_default=True,
        ),
    ] = 10,
    permutations: Annotated[
        int,
        Parameter(
            group="Model Parameters",
            show_default=True,
        ),
    ] = 1000,
    scoring: Annotated[
        ScoringMethod,
        Parameter(
            group="Model Parameters",
            show_default=True,
        ),
    ] = ScoringMethod.r2,
    processes: Annotated[
        int,
        Parameter(
            group="Model Parameters",
            show_default=True,
        ),
    ] = 1,
    plot_distributions: Annotated[
        bool,
        Parameter(
            show_default=True,
            group="Visualization Options",
        ),
    ] = False,
    verbose: Annotated[
        bool,
        Parameter(
            "-v",
            "--verbose",
            group="Optional parameters",
        ),
    ] = False,
    save_parameters: Annotated[
        bool,
        Parameter(
            "-s",
            "--save_parameters",
            group="Optional parameters",
        ),
    ] = False,
    overwrite: Annotated[
        bool,
        Parameter(
            "-f",
            "--overwrite",
            group="Optional parameters",
        ),
    ] = False
):
    """PARTIAL LEAST SQUARE REGRESSION (PLSR)
    --------------------------------------
    Plsr performs a Partial Least Square Regression (PLSR) on a graph
    using the nodes' attributes as predictors and the edges' weights as
    response variable. The script will perform a cross-validation to determine
    the optimal number of components to use for the PLSR model. It will then
    perform a permutation testing to determine if the model is statistically
    significant. Finally, it will output the PLSR coefficients and statistics
    as well as plots of the distributions of the attributes and edges' weights
    and the PLSR coefficients.

    PREPROCESSING
    -------------
    The script will scale the data to unit variance and zero mean and will
    perform a log transformation on the edges' weights (for now, it assumes
    that the weights represent a membership value resulting from a fuzzy
    clustering analysis).

    NODES' ATTRIBUTES
    -----------------
    The script takes only one graph file as input. The graph file must be in
    .gexf format. The script will then fetch the attributes from the graph
    file and will perform the PLSR analysis on the attributes and edges'
    weights. If no attributes are provided, the script will use all attributes
    found in the graph file. To set attributes to the nodes in the graph file,
    please see AddNodesAttributes.

    SCORING OPTIONS
    ---------------
    The script will perform a permutation testing to determine if the model is
    statistically significant. The script will compute the p-value for the
    permutation testing using the R2 score by default. However, the user can
    also choose multiple scores to compute the p-value. The available scores
    can be seen in [1]. The equation used to compute the single-tailed p-value
    is:

        p-value = ∑(score_perm >= score) / (nb_permutations)

    COEFFICIENT SIGNIFICANCE
    ------------------------
    The script will also compute the p-value for the coefficients using the
    permutation testing. The p-value for the coefficients is computed by
    comparing the coefficients obtained from the PLSR model with the
    coefficients obtained from the permutation testing. The equation used to
    compute the two-tailed p-value is:

        p-value = ∑(abs(coef_perm) >= abs(coef)) / (nb_permutations)

    REFERENCES
    ----------
    [1] [Scikit-learn scoring
    methods](https://scikit-learn.org/stable/modules/model_evaluation.html#scoring-parameter)

    EXAMPLE USAGE
    -------------
    ::

        PartialLeastSquareRegression --in-graph graph.gexf
        --out-folder output_folder -v -s

    Parameters
    ----------
    in_graph : str
        Graph file containing the data for the PLSR model.
    out_folder : str
        Output folder.
    attributes : List[str], optional
        Attributes names to include in the PLSR model. If None are provided,
        all attributes will be included.
    weight : str, optional
        Edge weight to use for the PLSR model.
    splits : int, optional
        Number of splits to use for the cross-validation.
    permutations : int, optional
        Number of permutations to use for the permutation testing.
    scoring : ScoringMethod, optional
        Scoring method to use for the permutation testing.
    processes : int, optional
        Number of processes to use for the cross-validation.
    verbose : bool, optional
        If true, produce verbose output.
    plot_distributions : bool, optional
        If true, will plot the distributions of the edges' weights.
    save_parameters : bool, optional
        If true, will save input parameters to .txt file.
    overwrite : bool, optional
        If true, force overwriting of existing output files.
    """

    warnings.filterwarnings("ignore")

    if verbose:
        logging.getLogger().setLevel(logging.INFO)
        coloredlogs.install(level=logging.INFO)

    assert_input(in_graph)
    assert_output_dir_exist(overwrite, out_folder, create_dir=True)

    # Creating output structure.
    os.makedirs(f"{out_folder}/Distributions", exist_ok=True)
    os.makedirs(f"{out_folder}/Coefficients", exist_ok=True)

    # Saving parameters
    if save_parameters:
        parameters = list(locals().items())
        with open(f"{out_folder}/nodes_attributes_parameters.txt", "w+") as f:
            for param in parameters:
                f.writelines(str(param))

    # Load graph file.
    logging.info("Loading graph and dataset.")
    G = nx.read_gml(in_graph)

    # Fetching dataframe.
    logging.info("Fetching nodes' attributes dataframe.")
    attr_df = fetch_attributes_df(G, attributes)

    # Fetching edge data.
    logging.info("Fetching edge data.")
    edge_df = fetch_edge_data(G, weight=weight)

    # Scaling data to unit variance and zero mean.
    logging.info("Scaling data.")
    attr = scale(attr_df)
    edge = edge_df.apply(lambda x: np.log(x))
    edge = scale(edge)

    # Performing Cross-Validation.
    logging.info("Performing Cross-Validation.")
    (plsr, mse, score_c, score_cv,
     rscore, mse_c, mse_cv) = plsr_cv(
        attr,
        edge,
        len(attr_df.columns),
        splits=splits,
        processes=processes,
        verbose=verbose)

    # Permutation testing.
    logging.info("Performing permutation testing.")
    mod, score, coef, perm_score, score_pval, perm_coef, coef_pval = \
        permutation_testing(
            plsr,
            attr,
            edge,
            nb_permutations=permutations,
            scoring=scoring,
            splits=splits,
            processes=processes,
            verbose=verbose)

    # Exporting statistics.
    logging.info("Exporting statistics.")
    coef = {
        f'coef{i+1}': coef[:, i] for i in range(0, coef.shape[0])}
    coef['varname'] = attr_df.columns
    coef_df = pd.DataFrame(coef)
    coef_df.to_excel(f"{out_folder}/Coefficients/coef_df.xlsx")

    coef_pval_df = pd.DataFrame(coef_pval,
                                index=attr_df.columns,
                                columns=edge_df.columns)
    coef_pval_df.to_excel(f"{out_folder}/Coefficients/coef_pval_df.xlsx")

    stats = pd.DataFrame([mse_c, mse_cv, score_c, score_cv, rscore,
                          score, perm_score, score_pval],
                         columns=['Statistics'],
                         index=['MSE_c', 'MSE_cv', 'R2_c', 'R2_cv', 'R_c',
                                'R2_score', 'R2_perm', 'pval'])
    stats.to_excel(f"{out_folder}/statistics.xlsx", header=True, index=True)

    # Plotting results.
    logging.info("Generating plots.")

    # Plotting histograms.
    with plt.rc_context(
        {"font.size": 10, "font.weight": "bold", "axes.titleweight": "bold"}
    ):
        if plot_distributions:
            # Edge's distributions.
            edge_viz = pd.DataFrame(edge, columns=edge_df.columns)
            flexible_hist(edge_viz,
                          f"{out_folder}/Distributions/edges_distribution.png",
                          cmap="magma", title="Edges' distributions",
                          xlabel="Edges' weights", ylabel="Density")

        # MSE plot.
        fig, ax = plt.subplots(figsize=(10, 10))
        plt.plot(
            np.arange(1, len(attr_df.columns) + 1),
            np.array(mse),
            '-v',
            color='blue',
            mfc='blue')
        plt.plot(
            np.arange(1, len(attr_df.columns) + 1)[np.argmin(mse)],
            np.array(mse)[np.argmin(mse)],
            'P',
            ms=10,
            mfc='red')
        plt.xlabel('Number of PLS components')
        plt.ylabel('MSE')
        plt.title('MSE plot for multiple PLS components number.')
        plt.xlim(left=-1)
        plt.tight_layout()
        plt.savefig(f"{out_folder}/mse_plot.png")
        plt.close()

        # Coefficient plot.
        for i in range(0, len(edge_df.columns)):
            generate_coef_plot(
                coef_df,
                coef_pval[:, i],
                coefname=f'coef{i+1}',
                varname='varname',
                output=f"{out_folder}/Coefficients/coef_plot_cluster_{i+1}.png"
            )


if __name__ == "__main__":
    app()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import coloredlogs
import logging

from cyclopts import App, Parameter
import networkx as nx
import numpy as np
from typing_extensions import Annotated

from neurostatx.io.utils import assert_input, assert_output_dir_exist
from neurostatx.network.utils import (extract_subject_percentile,
                                      fetch_edge_data)
from neurostatx.network.viz import (visualize_network,
                                    creating_node_colormap)


# Initializing the app.
app = App(default_parameter=Parameter(negative=()))


@app.default()
def CompareGraphs(
    in_graph1: Annotated[
        str,
        Parameter(
            show_default=False,
            group="Essential Files Options",
        ),
    ],
    weight: Annotated[
        str,
        Parameter(
            show_default=False,
            group="Essential Files Options",
        ),
    ],
    percentile: Annotated[
        float,
        Parameter(
            show_default=False,
            group="Essential Files Options",
        ),
    ],
    in_graph2: Annotated[
        str,
        Parameter(
            show_default=False,
            group="Essential Files Options",
        ),
    ],
    out_folder: Annotated[
        str,
        Parameter(
            group="Essential Files Options",
        ),
    ] = "./comparison_results/",
    verbose: Annotated[
        bool,
        Parameter(
            "-v",
            "--verbose",
            group="Optional parameters",
        ),
    ] = False,
    overwrite: Annotated[
        bool,
        Parameter(
            "-f",
            "--overwrite",
            group="Optional parameters",
        ),
    ] = False,
    label_centroids: Annotated[
        bool,
        Parameter(
            show_default=True,
            group="Network Visualization Options",
        ),
    ] = True,
    label_subjects: Annotated[
        bool,
        Parameter(
            show_default=True,
            group="Network Visualization Options",
        ),
    ] = False,
    centroids_size: Annotated[
        int,
        Parameter(
            show_default=True,
            group="Network Visualization Options",
        ),
    ] = 500,
    centroid_alpha: Annotated[
        float,
        Parameter(
            show_default=True,
            group="Network Visualization Options",
        ),
    ] = 1,
    centroid_node_color: Annotated[
        str,
        Parameter(
            show_default=True,
            group="Network Visualization Options",
        ),
    ] = "white",
    centroid_edge_color: Annotated[
        str,
        Parameter(
            show_default=True,
            group="Network Visualization Options",
        ),
    ] = "black",
    subject_node_size: Annotated[
        int,
        Parameter(
            show_default=True,
            group="Network Visualization Options",
        ),
    ] = 5,
    background_alpha: Annotated[
        bool,
        Parameter(
            show_default=True,
            group="Network Visualization Options",
        ),
    ] = True,
    subject_edge_color: Annotated[
        str,
        Parameter(
            show_default=True,
            group="Network Visualization Options",
        ),
    ] = None,
    colormap: Annotated[
        str,
        Parameter(
            show_default=True,
            group="Network Visualization Options",
        ),
    ] = "gray",
    legend_title: Annotated[
        str,
        Parameter(
            show_default=False,
            group="Network Visualization Options",
        ),
    ] = "Membership values",
):
    """GRAPH NETWORK COMPARISON
    --------------------------------------
    CompareGraphs is a script that compares 2 undirected weighted
    graph network. As of now, the only comparison implemented is the
    extraction of the Xth percentile nodes from --in-graph1 and label those
    nodes on --in-graph2. The percentile value is set by --percentile.

    GRAPH NETWORK CUSTOMIZATION
    ---------------------------
    To customize the graph appearance, please see the Network Visualization
    Options below. It should be noted that using subjects_labelling will crowd
    the network if it contains a high number of nodes. Also, centroids are
    labelled by default 'c1, c2, ...' and subjects 's1, s2, ...'.

    EXAMPLE USAGE
    -------------
    ::

        CompareGraphs --in-graph1 graph1.gml --in-matrix membership_mat.npy
        --percentile 80 --in-graph2 graph2.gml

    Parameters
    ----------
    in_graph1 : str
        1st graph from which subjects above --percentile will be extracted and
        colored.
    weight : str
        Edge's weight to use for the graph.
    percentile : float
        Percentile value used to extract subjects.
    in_graph2 : str
        2nd graph to color extracted subjects on.
    out_folder : str, optional
        Path of the folder in which the results will be written. If not
        specified, current folder and default name will be used.
    verbose : bool, optional
        If true, produce verbose output.
    overwrite : bool, optional
        If true, force overwriting of existing output files.
    label_centroids : bool, optional
        If true, centroids will be labelled.
    label_subjects : bool, optional
        If true, will label subjects nodes.
    centroids_size : int, optional
        Size of the centroids nodes.
    centroid_alpha : float, optional
        Alpha value representing the transparency of the centroids nodes.
    centroid_node_color : str, optional
        Centroids nodes color to use.
    centroid_edge_color : str, optional
        Assign a color to the edge of the centroids nodes.
    subject_node_size : int, optional
        Assign the size of the subjects nodes.
    background_alpha : bool, optional
        If true, will plot the background nodes with alpha = 0.2.
    subject_edge_color : str, optional
        Assign a color to the edge of the subjects nodes.
    colormap : str, optional
        Colormap to use when coloring the edges of the network based on the
        membership values to each clusters. Available colormap are those from
        [Matplotlib](https://matplotlib.org/stable/tutorials/colors/colormaps.html).
    legend_title : str, optional
        Legend title (colormap).
    """

    if verbose:
        logging.getLogger().setLevel(logging.INFO)
        coloredlogs.install(level=logging.INFO)

    assert_input(in_graph1, in_graph2)
    assert_output_dir_exist(overwrite, out_folder, create_dir=True)

    # Loading membership matrix.
    logging.info("Loading graphs.")
    graph1 = nx.read_gml(in_graph1)
    graph2 = nx.read_gml(in_graph2)

    # Fetch data from graph #1.
    mat = fetch_edge_data(graph1, weight=weight)

    # Extracting percentiles.
    logging.info("Extracting percentiles.")
    # Extracting the Xth percentile subjects.
    percentile_dict = extract_subject_percentile(mat.values.T, percentile)

    # Mapping the nodes' cmap.
    nodes_cmap = creating_node_colormap(percentile_dict)

    # Creating the alpha for subject's that are not in the Xth percentile.
    if background_alpha:
        sub_alpha = []
        for i in nodes_cmap:
            if i is str:
                sub_alpha.append(0.2)
            else:
                sub_alpha.append(1)
    else:
        sub_alpha = np.array([1] * mat.values.shape[0])

    logging.info("Visualizing percentiles on the 1st graph.")
    visualize_network(
        graph1,
        output=f"{out_folder}/graph1.png",
        weight="membership",
        centroids_labelling=label_centroids,
        subjects_labelling=label_subjects,
        centroid_node_shape=centroids_size,
        centroid_alpha=centroid_alpha,
        centroid_node_color=centroid_node_color,
        centroid_edge_color=centroid_edge_color,
        subject_node_shape=subject_node_size,
        subject_alpha=sub_alpha,
        subject_node_color=nodes_cmap,
        subject_edge_color=subject_edge_color,
        colormap=colormap,
        title="Graph Network #1",
        legend_title=legend_title,
    )

    logging.info("Visualizing percentiles on the 2nd graph.")
    visualize_network(
        graph2,
        output=f"{out_folder}/graph2.png",
        weight="membership",
        centroids_labelling=label_centroids,
        subjects_labelling=label_subjects,
        centroid_node_shape=centroids_size,
        centroid_alpha=centroid_alpha,
        centroid_node_color=centroid_node_color,
        centroid_edge_color=centroid_edge_color,
        subject_node_shape=subject_node_size,
        subject_alpha=sub_alpha,
        subject_node_color=nodes_cmap,
        subject_edge_color=subject_edge_color,
        colormap=colormap,
        title="Graph Network #2",
        legend_title=legend_title,
    )


if __name__ == "__main__":
    app()



# Originally written by Nick Cullen
# Extended and currently maintained by JP Fortin
# This file is part of the neuroCombat package and has been modified to
# support recent numpy update used within NeuroStatX.
# All credits go to the original authors.
# https://github.com/Jfortin1/neuroCombat
# Will be removed once neuroCombat is updated to support recent numpy version.
import logging as log

import pandas as pd
import numpy as np
import numpy.linalg as la
import math
import copy


def neuroCombat(
    dat,
    covars,
    batch_col,
    categorical_cols=None,
    continuous_cols=None,
    eb=True,
    parametric=True,
    mean_only=False,
    ref_batch=None,
):
    """
    Run ComBat to remove scanner effects in multi-site imaging data

    Arguments
    ---------
    dat : a pandas data frame or numpy array
        - neuroimaging data to correct with shape = (features, samples) e.g.
        cortical thickness measurements, image voxels, etc

    covars : a pandas data frame w/ shape = (samples, covariates)
        - contains the batch/scanner covariate as well as additional
        covariates (optional) that should be preserved during harmonization.

    batch_col : string
        - indicates batch (scanner) column name in covars (e.g. "scanner")

    categorical_cols : list of strings
        - specifies column names in covars data frame of categorical variables
        to be preserved during harmonization (e.g. ["sex", "disease"])

    continuous_cols : list of strings
        - indicates column names in covars data frame of continuous variables
        to be preserved during harmonization (e.g. ["age"])

    eb : should Empirical Bayes be performed?
        - True by default

    parametric : should parametric adjustements be performed?
        - True by default

    mean_only : should only be the mean adjusted (no scaling)?
        - False by default

    ref_batch : batch (site or scanner) to be used as reference for batch
        adjustment.
        - None by default

    Returns
    -------
    A dictionary of length 3:
    - data: A numpy array with the same shape as `dat` which has now been
        ComBat-harmonized
    - estimates: A dictionary of the ComBat estimates used for harmonization
    - info: A dictionary of the inputs needed for ComBat harmonization
    """
    ##############################
    # CLEANING UP INPUT DATA #
    ##############################
    if not isinstance(covars, pd.DataFrame):
        raise ValueError(
            "covars must be pandas dataframe -> try: covars = \
                        pandas.DataFrame(covars)"
        )

    if not isinstance(categorical_cols, (list, tuple)):
        if categorical_cols is None:
            categorical_cols = []
        else:
            categorical_cols = [categorical_cols]
    if not isinstance(continuous_cols, (list, tuple)):
        if continuous_cols is None:
            continuous_cols = []
        else:
            continuous_cols = [continuous_cols]

    covar_labels = np.array(covars.columns)
    covars = np.array(covars, dtype="object")

    if isinstance(dat, pd.DataFrame):
        dat = np.array(dat, dtype="float32")

    ##############################

    # get column indices for relevant variables
    batch_col = np.where(covar_labels == batch_col)[0][0]
    cat_cols = [np.where(covar_labels == c_var)[0][0] for c_var in
                categorical_cols]
    num_cols = [np.where(covar_labels == n_var)[0][0] for n_var in
                continuous_cols]

    # convert batch col to integer
    if ref_batch is None:
        ref_level = None
    else:
        ref_indices = np.argwhere(covars[:, batch_col] == ref_batch).squeeze()
        if ref_indices.shape[0] == 0:
            ref_level = None
            ref_batch = None
            log.info("[neuroCombat] batch.ref not found. Setting to None.")
            covars[:, batch_col] = np.unique(covars[:, batch_col],
                                             return_inverse=True)[
                -1
            ]
        else:
            covars[:, batch_col] = np.unique(covars[:, batch_col],
                                             return_inverse=True)[
                -1
            ]
            ref_level = covars[int(ref_indices[0]), batch_col]
    # create dictionary that stores batch info
    (batch_levels, sample_per_batch) = np.unique(
        covars[:, batch_col], return_counts=True
    )

    # create design matrix
    ("[neuroCombat] Creating design matrix")
    design = make_design_matrix(covars, batch_col, cat_cols, num_cols,
                                ref_level)

    info_dict = {
        "batch_levels": batch_levels,
        "ref_level": ref_level,
        "n_batch": len(batch_levels),
        "n_sample": int(covars.shape[0]),
        "sample_per_batch": sample_per_batch.astype("int"),
        "batch_info": [
            list(np.where(covars[:, batch_col] == idx)[0]) for idx in
            batch_levels
        ],
        "design": design,
    }

    # standardize data across features
    log.info("[neuroCombat] Standardizing data across features")
    s_data, s_mean, v_pool, mod_mean = standardize_across_features(
        dat, design, info_dict
    )

    # fit L/S models and find priors
    log.info("[neuroCombat] Fitting L/S model and finding priors")
    LS_dict = fit_LS_model_and_find_priors(s_data, design, info_dict,
                                           mean_only)

    # find parametric adjustments
    if eb:
        if parametric:
            log.info("[neuroCombat] Finding parametric adjustments")
            gamma_star, delta_star = find_parametric_adjustments(
                s_data, LS_dict, info_dict, mean_only
            )
        else:
            log.info("[neuroCombat] Finding non-parametric adjustments")
            gamma_star, delta_star = find_non_parametric_adjustments(
                s_data, LS_dict, info_dict, mean_only
            )
    else:
        log.info("[neuroCombat] Finding L/S adjustments without Empirical \
            Bayes")
        gamma_star, delta_star = find_non_eb_adjustments(s_data, LS_dict,
                                                         info_dict)

    # adjust data
    log.info("[neuroCombat] Final adjustment of data")
    bayes_data = adjust_data_final(
        s_data, design, gamma_star, delta_star, s_mean, mod_mean, v_pool,
        info_dict, dat
    )

    bayes_data = np.array(bayes_data)
    estimates = {
        "batches": info_dict["batch_levels"],
        "var.pooled": v_pool,
        "stand.mean": s_mean,
        "mod.mean": mod_mean,
        "gamma.star": gamma_star,
        "delta.star": delta_star,
    }
    estimates = {
        **LS_dict,
        **estimates,
    }

    return {"data": bayes_data, "estimates": estimates, "info": info_dict}


def make_design_matrix(Y, batch_col, cat_cols, num_cols, ref_level):
    """
    Return Matrix containing the following parts:
        - one-hot matrix of batch variable (full)
        - one-hot matrix for each categorical_cols (removing the first column)
        - column for each continuous_cols
    """

    def to_categorical(y, nb_classes=None):
        if not nb_classes:
            nb_classes = np.max(y) + 1
        Y = np.zeros((len(y), nb_classes))
        for i in range(len(y)):
            Y[i, y[i]] = 1.0
        return Y

    hstack_list = []

    # batch one-hot #
    # convert batch column to integer in case it's string
    batch = np.unique(Y[:, batch_col], return_inverse=True)[-1]
    batch_onehot = to_categorical(batch, len(np.unique(batch)))
    if ref_level is not None:
        batch_onehot[:, ref_level] = np.ones(batch_onehot.shape[0])
    hstack_list.append(batch_onehot)

    # categorical one-hots #
    for cat_col in cat_cols:
        cat = np.unique(np.array(Y[:, cat_col]), return_inverse=True)[1]
        cat_onehot = to_categorical(cat, len(np.unique(cat)))[:, 1:]
        hstack_list.append(cat_onehot)

    # numerical vectors #
    for num_col in num_cols:
        num = np.array(Y[:, num_col], dtype="float32")
        num = num.reshape(num.shape[0], 1)
        hstack_list.append(num)

    design = np.hstack(hstack_list)
    return design


def standardize_across_features(X, design, info_dict):
    n_batch = info_dict["n_batch"]
    n_sample = info_dict["n_sample"]
    sample_per_batch = info_dict["sample_per_batch"]
    batch_info = info_dict["batch_info"]
    ref_level = info_dict["ref_level"]

    def get_beta_with_nan(yy, mod):
        wh = np.isfinite(yy)
        mod = mod[wh, :]
        yy = yy[wh]
        B = np.dot(np.dot(la.inv(np.dot(mod.T, mod)), mod.T), yy.T)
        return B

    betas = []
    for i in range(X.shape[0]):
        betas.append(get_beta_with_nan(X[i, :], design))
    B_hat = np.vstack(betas).T

    # B_hat = np.dot(np.dot(la.inv(np.dot(design.T, design)), design.T), X.T)
    if ref_level is not None:
        grand_mean = np.transpose(B_hat[ref_level, :])
    else:
        grand_mean = np.dot((sample_per_batch / float(n_sample)).T,
                            B_hat[:n_batch, :])
    stand_mean = np.dot(
        grand_mean.T.reshape((len(grand_mean), 1)), np.ones((1, n_sample))
    )
    # var_pooled = np.dot(((X - np.dot(design, B_hat).T)**2),
    # np.ones((n_sample, 1)) / float(n_sample))

    if ref_level is not None:
        X_ref = X[:, batch_info[ref_level]]
        design_ref = design[batch_info[ref_level], :]
        n_sample_ref = sample_per_batch[ref_level]
        var_pooled = np.dot(
            ((X_ref - np.dot(design_ref, B_hat).T) ** 2),
            np.ones((n_sample_ref, 1)) / float(n_sample_ref),
        )
    else:
        var_pooled = np.dot(
            ((X - np.dot(design, B_hat).T) ** 2),
            np.ones((n_sample, 1)) / float(n_sample),
        )

    var_pooled[var_pooled == 0] = np.median(var_pooled != 0)

    mod_mean = 0
    if design is not None:
        tmp = copy.deepcopy(design)
        tmp[:, range(0, n_batch)] = 0
        mod_mean = np.transpose(np.dot(tmp, B_hat))
    # Continue here.

    # tmp = np.array(design.copy())
    # tmp[:,:n_batch] = 0
    # stand_mean  += np.dot(tmp, B_hat).T

    s_data = (X - stand_mean - mod_mean) / np.dot(
        np.sqrt(var_pooled), np.ones((1, n_sample))
    )

    return s_data, stand_mean, var_pooled, mod_mean


def aprior(delta_hat):
    m = np.mean(delta_hat)
    s2 = np.var(delta_hat, ddof=1)
    return (2 * s2 + m**2) / float(s2)


def bprior(delta_hat):
    m = delta_hat.mean()
    s2 = np.var(delta_hat, ddof=1)
    return (m * s2 + m**3) / s2


def postmean(g_hat, g_bar, n, d_star, t2):
    return (t2 * n * g_hat + d_star * g_bar) / (t2 * n + d_star)


def postvar(sum2, n, a, b):
    return (0.5 * sum2 + b) / (n / 2.0 + a - 1.0)


def convert_zeroes(x):
    x[x == 0] = 1
    return x


def fit_LS_model_and_find_priors(s_data, design, info_dict, mean_only):
    n_batch = info_dict["n_batch"]
    batch_info = info_dict["batch_info"]

    batch_design = design[:, :n_batch]
    gamma_hat = np.dot(
        np.dot(la.inv(np.dot(batch_design.T, batch_design)), batch_design.T),
        s_data.T
    )

    delta_hat = []
    for i, batch_idxs in enumerate(batch_info):
        if mean_only:
            delta_hat.append(np.repeat(1, s_data.shape[0]))
        else:
            delta_hat.append(np.var(s_data[:, batch_idxs], axis=1, ddof=1))

    delta_hat = list(map(convert_zeroes, delta_hat))
    gamma_bar = np.mean(gamma_hat, axis=1)
    t2 = np.var(gamma_hat, axis=1, ddof=1)

    if mean_only:
        a_prior = None
        b_prior = None
    else:
        a_prior = list(map(aprior, delta_hat))
        b_prior = list(map(bprior, delta_hat))

    LS_dict = {}
    LS_dict["gamma_hat"] = gamma_hat
    LS_dict["delta_hat"] = delta_hat
    LS_dict["gamma_bar"] = gamma_bar
    LS_dict["t2"] = t2
    LS_dict["a_prior"] = a_prior
    LS_dict["b_prior"] = b_prior
    return LS_dict


# Helper function for parametric adjustements:
def it_sol(sdat, g_hat, d_hat, g_bar, t2, a, b, conv=0.0001):
    n = (1 - np.isnan(sdat)).sum(axis=1)
    g_old = g_hat.copy()
    d_old = d_hat.copy()

    change = 1
    count = 0
    while change > conv:
        g_new = postmean(g_hat, g_bar, n, d_old, t2)
        sum2 = (
            (
                sdat - np.dot(
                    g_new.reshape((g_new.shape[0], 1)), np.ones((1,
                                                                 sdat.shape[1])
                                                                )
                )
            ) ** 2
        ).sum(axis=1)
        d_new = postvar(sum2, n, a, b)

        change = max(
            (abs(g_new - g_old) / g_old).max(),
            (abs(d_new - d_old) / d_old).max()
        )
        g_old = g_new  # .copy()
        d_old = d_new  # .copy()
        count = count + 1
    adjust = (g_new, d_new)
    return adjust


# Helper function for non-parametric adjustements:
def int_eprior(sdat, g_hat, d_hat):
    r = sdat.shape[0]
    gamma_star, delta_star = [], []
    for i in range(0, r, 1):
        g = np.delete(g_hat, i)
        d = np.delete(d_hat, i)
        x = sdat[i, :]
        n = x.shape[0]
        j = np.repeat(1, n)
        A = np.repeat(x, g.shape[0])
        A = A.reshape(n, g.shape[0])
        A = np.transpose(A)
        B = np.repeat(g, n)
        B = B.reshape(g.shape[0], n)
        resid2 = np.square(A - B)
        sum2 = resid2.dot(j)
        LH = 1 / (2 * math.pi * d) ** (n / 2) * np.exp(-sum2 / (2 * d))
        LH = np.nan_to_num(LH)
        gamma_star.append(sum(g * LH) / sum(LH))
        delta_star.append(sum(d * LH) / sum(LH))
    adjust = (gamma_star, delta_star)
    return adjust


def find_parametric_adjustments(s_data, LS, info_dict, mean_only):
    batch_info = info_dict["batch_info"]
    ref_level = info_dict["ref_level"]

    gamma_star, delta_star = [], []
    for i, batch_idxs in enumerate(batch_info):
        if mean_only:
            gamma_star.append(
                postmean(LS["gamma_hat"][i], LS["gamma_bar"][i], 1, 1,
                         LS["t2"][i])
            )
            delta_star.append(np.repeat(1, s_data.shape[0]))
        else:
            temp = it_sol(
                s_data[:, batch_idxs],
                LS["gamma_hat"][i],
                LS["delta_hat"][i],
                LS["gamma_bar"][i],
                LS["t2"][i],
                LS["a_prior"][i],
                LS["b_prior"][i],
            )
            gamma_star.append(temp[0])
            delta_star.append(temp[1])

    gamma_star = np.array(gamma_star)
    delta_star = np.array(delta_star)

    if ref_level is not None:
        gamma_star[ref_level, :] = np.zeros(gamma_star.shape[-1])
        delta_star[ref_level, :] = np.ones(delta_star.shape[-1])

    return gamma_star, delta_star


def find_non_parametric_adjustments(s_data, LS, info_dict, mean_only):
    batch_info = info_dict["batch_info"]
    ref_level = info_dict["ref_level"]

    gamma_star, delta_star = [], []
    for i, batch_idxs in enumerate(batch_info):
        if mean_only:
            LS["delta_hat"][i] = np.repeat(1, s_data.shape[0])
        temp = int_eprior(s_data[:, batch_idxs], LS["gamma_hat"][i],
                          LS["delta_hat"][i])

        gamma_star.append(temp[0])
        delta_star.append(temp[1])

    gamma_star = np.array(gamma_star)
    delta_star = np.array(delta_star)

    if ref_level is not None:
        gamma_star[ref_level, :] = np.zeros(gamma_star.shape[-1])
        delta_star[ref_level, :] = np.ones(delta_star.shape[-1])

    return gamma_star, delta_star


def find_non_eb_adjustments(s_data, LS, info_dict):
    gamma_star = np.array(LS["gamma_hat"])
    delta_star = np.array(LS["delta_hat"])
    ref_level = info_dict["ref_level"]

    if ref_level is not None:
        gamma_star[ref_level, :] = np.zeros(gamma_star.shape[-1])
        delta_star[ref_level, :] = np.ones(delta_star.shape[-1])

    return gamma_star, delta_star


def adjust_data_final(
    s_data,
    design,
    gamma_star,
    delta_star,
    stand_mean,
    mod_mean,
    var_pooled,
    info_dict,
    dat,
):
    sample_per_batch = info_dict["sample_per_batch"]
    n_batch = info_dict["n_batch"]
    n_sample = info_dict["n_sample"]
    batch_info = info_dict["batch_info"]
    ref_level = info_dict["ref_level"]

    batch_design = design[:, :n_batch]

    bayesdata = s_data
    gamma_star = np.array(gamma_star)
    delta_star = np.array(delta_star)

    for j, batch_idxs in enumerate(batch_info):
        dsq = np.sqrt(delta_star[j, :])
        dsq = dsq.reshape((len(dsq), 1))
        denom = np.dot(dsq, np.ones((1, sample_per_batch[j])))
        numer = np.array(
            bayesdata[:, batch_idxs] - np.dot(batch_design[batch_idxs, :],
                                              gamma_star).T
        )

        bayesdata[:, batch_idxs] = numer / denom

    vpsq = np.sqrt(var_pooled).reshape((len(var_pooled), 1))
    bayesdata = bayesdata * np.dot(vpsq, np.ones((1, n_sample))) + stand_mean \
        + mod_mean

    if ref_level is not None:
        bayesdata[:, batch_info[ref_level]] = dat[:, batch_info[ref_level]]

    return bayesdata



import unittest
import numpy as np
from neurostatx.clustering.fuzzy import fuzzyCmeans


class TestFuzzyCmeans(unittest.TestCase):

    def setUp(self):
        # Define three centroids
        self.centroids = np.array([[1, 1], [3, 3], [5, 5]])

        # Generate data around the centroids
        np.random.seed(0)
        self.X = np.concatenate([
            self.centroids[i] + 0.2 * np.random.randn(200, 2)
            for i in range(len(self.centroids))
        ])

    def test_fuzzyCmeans(self):
        # Test the function with generated data
        cntr, _, _, _, _, _, _, _, _ = fuzzyCmeans(self.X, max_cluster=3)

        # Check if the returned centroids are close to the initially chosen
        # centroids
        for i in range(len(self.centroids)):
            self.assertTrue(np.allclose(np.sort(cntr[1], axis=0)[i],
                                        self.centroids[i],
                                        atol=0.1))


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import coloredlogs
import logging
import os

from cyclopts import App, Parameter
import networkx as nx
from typing import List
from typing_extensions import Annotated

from neurostatx.io.utils import (assert_input, assert_output,
                                 load_df_in_any_format)
from neurostatx.network.utils import construct_attributes_dict

# Initializing the app.
app = App(default_parameter=Parameter(negative=()))


@app.default()
def AddNodesAttributes(
    in_graph: Annotated[
        str,
        Parameter(
            show_default=False,
            group="Essential Files Options",
        ),
    ],
    in_dataset: Annotated[
        str,
        Parameter(
            show_default=False,
            group="Essential Files Options",
        ),
    ],
    id_column: Annotated[
        str,
        Parameter(
            show_default=False,
            group="Essential Files Options",
        ),
    ],
    labels: Annotated[
        List[str],
        Parameter(
            show_default=False,
            group="Essential Files Options",
        ),
    ],
    out_file: Annotated[
        str,
        Parameter(
            show_default=False,
            group="Essential Files Options",
        ),
    ] = "graph_with_attributes.gexf",
    verbose: Annotated[
        bool,
        Parameter(
            "-v",
            "--verbose",
            group="Optional parameters",
        ),
    ] = False,
    overwrite: Annotated[
        bool,
        Parameter(
            "-f",
            "--overwrite",
            group="Optional parameters",
        ),
    ] = False
):
    """SETTING NODES ATTRIBUTES
    ------------------------
    AddNodesAttributes is a script that sets the attributes of the
    nodes of a graph. The attributes are provided via a tabulated data format
    (.csv, .xlsx or .txt) file. The script will automatically match the IDs of
    the nodes with the IDs of the subjects in the file. Name of the variables
    to add as attributes can be supplied either via a .txt file or with
    multiple --labels arguments.

    EXAMPLE USAGE
    -------------
    ::

        AddNodesAttributes --in-graph graph.gml --in-dataset dataset.xlsx
        --id-column ID --labels label1 --labels label2 --labels label3
        --out-file graph_attributes.gexf --verbose

    Parameters
    ----------
    in_graph : str
        Graph file to add attributes to.
    in_dataset : str
        Dataset containing the variables to add as nodes' attributes to the
        graph.
    id_column : str
        Name of the column containing the subject's ID tag. Required for proper
        handling of IDs and merging multiple datasets.
    labels : List[str]
        Label(s) name(s) to add as nodes' attributes to the graph. Can be
        supplied multiple times or as a .txt file containing all names in a
        line separated format.
    out_file : str, optional
        Output graph file name (.gml).
    verbose : bool, optional
        If true, produce verbose output.
    overwrite : bool, optional
        If true, force overwriting of existing output files.
    """

    if verbose:
        logging.getLogger().setLevel(logging.INFO)
        coloredlogs.install(level=logging.INFO)

    assert_input(in_graph)
    assert_input(in_dataset)
    assert_output(overwrite, out_file)

    logging.info("Loading graph and dataset.")
    G = nx.read_gml(in_graph)
    df = load_df_in_any_format(in_dataset)

    # Sorting if labels is a .txt file or not.
    if len(labels) == 1:
        filename, ext = os.path.splitext(labels[0])
        if ext == '.txt':
            print('yeah')
            with open(labels[0], 'r') as f:
                labels = f.read().splitlines()

    # Getting dictonary of labels and values to add.
    logging.info("Constructing dictionary of attributes to add.")
    attributes = construct_attributes_dict(df, labels, id_column)

    # Add attributes to graph.
    logging.info("Adding attributes to graph.")
    nx.set_node_attributes(G, attributes)

    logging.info("Saving graph.")
    nx.write_gml(G, out_file)


if __name__ == "__main__":
    app()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Import required libraries.
import coloredlogs
import dill as pickle
import logging
import sys

from cyclopts import App, Parameter
import pandas as pd
import semopy
from typing import List
from typing_extensions import Annotated

from neurostatx.io.utils import (
    assert_input,
    assert_output_dir_exist,
    load_df_in_any_format
)
from neurostatx.utils.preprocessing import merge_dataframes
from neurostatx.utils.factor import cfa


# Initializing the app.
app = App(default_parameter=Parameter(negative=()))


@app.default()
def ConfirmatoryFA(
    in_dataset: Annotated[
        List[str],
        Parameter(
            show_default=False,
            group="Essential Files Options",
        ),
    ],
    id_column: Annotated[
        str,
        Parameter(
            show_default=False,
            group="Essential Files Options",
        ),
    ],
    desc_columns: Annotated[
        int,
        Parameter(
            show_default=False,
            group="Essential Files Options",
        ),
    ],
    out_folder: Annotated[
        str,
        Parameter(
            group="Essential Files Options",
        ),
    ] = "./ResultsCFA/",
    loadings_df: Annotated[
        str,
        Parameter(
            show_default=False,
            group="Factorial Analysis parameters",
        ),
    ] = None,
    model: Annotated[
        List[str],
        Parameter(
            show_default=False,
            group="Factorial Analysis parameters",
        ),
    ] = None,
    threshold: Annotated[
        float,
        Parameter(
            "--threshold",
            group="Factorial Analysis parameters",
        ),
    ] = 0.40,
    iterations: Annotated[
        int,
        Parameter(
            "--iterations",
            group="Factorial Analysis parameters",
        ),
    ] = None,
    mean: Annotated[
        bool,
        Parameter(
            "--mean",
            group="Imputing parameters",
        ),
    ] = False,
    median: Annotated[
        bool,
        Parameter(
            "--median",
            group="Imputing parameters",
        ),
    ] = False,
    verbose: Annotated[
        bool,
        Parameter(
            "-v",
            "--verbose",
            group="Optional parameters",
        ),
    ] = False,
    save_parameters: Annotated[
        bool,
        Parameter(
            "-s",
            "--save_parameters",
            group="Optional parameters",
        ),
    ] = False,
    overwrite: Annotated[
        bool,
        Parameter(
            "-f",
            "--overwrite",
            group="Optional parameters",
        ),
    ] = False,
):
    """CONFIRMATORY FACTORIAL ANALYSIS
    -------------------------------
    ConfirmatoryFA is a script that can be used to perform a confirmatory
    factorial analysis (CFA) to test a hypothesized model of the relationships
    between observed variables and latent constructs. The script will output
    factor scores and statistics of goodness of fit such as Chi-square, RMSEA,
    CFI and TLI. The script will also generate a html report containing the
    results of the analysis. A good reference to understand those metrics can
    be accessed in [1].

    USING EFA SCORES OR CFA SCORES
    ------------------------------
    Both method can be used to derive factor scores. Since there is no clear
    consensus surrounding which is preferred (see [2]) the script will output
    both factor scores. As shown in [3], both methods highly correlate with one
    another. It then comes down to the user's preference.

    INPUT SPECIFICATIONS
    --------------------
    Dataset can contain multiple descriptive rows before the variables of
    interest. Simply specify the number of descriptive rows using
    --desc-columns. Rows with missing values will be removed by
    default, please select the mean or median option to impute missing data
    (be cautious when doing this).

    REFERENCES
    ----------
    [1] Costa, V., & Sarmento, R. Confirmatory Factor Analysis.
    https://arxiv.org/ftp/arxiv/papers/1905/1905.05598.pdf

    [2]
    https://stats.stackexchange.com/questions/346499/whether-to-use-efa-or-cfa-to-predict-latent-variables-scores

    [3] https://github.com/gagnonanthony/NeuroStatX/pull/11

    EXAMPLE USAGE
    -------------
    ::

        ConfirmatoryFA --in-dataset dataset.csv --id-column ID --desc-columns 1
        --out-folder ./output/ --loadings-df loadings.csv --threshold 0.40
        --mean -v -f

    Parameters
    ----------
    in_dataset : List[str]
        Input dataset(s) to use in the factorial analysis. If multiple files
        are provided as input, will be merged according to the subject id
        columns. For multiple inputs, use this: --in-dataset df1 --in-dataset
        df2 [...]
    id_column : str
        Name of the column containing the subject's ID tag. Required for proper
        handling of IDs and merging multiple datasets.
    desc_columns : int
        Number of descriptive columns at the beginning of the dataset to
        exclude in statistics and descriptive tables.
    out_folder : str, optional
        Path of the folder in which the results will be written. If not
        specified, current folder and default name will be used (e.g. =
        ./output/).
    loadings_df : str, optional
        Filename of the dataframe containing the loadings of the EFA analysis.
        Columns must be factors and rows variables.
    model : str, optional
        Model specification for the CFA analysis. Must be provided within
        brackets. (ex: --model "factor1 =~ var1 + var2 + var3"
        --model "factor2 =~ var4 + var5")
    threshold : float, optional
        Threshold to use to determine variables to include for each factor
        in CFA analysis. (ex: if set to 0.40, only variables with loadings
        higher than 0.40 will be assigned to a factor in the CFA model).
    iterations : int, optional
        Number of iterations to perform the bootstrapping of the model.
    mean : bool, optional
        Impute missing values in the original dataset based on the column mean.
    median : bool, optional
        Impute missing values in the original dataset based on the column
        median.
    verbose : bool, optional
        If true, produce verbose output.
    save_parameters : bool, optional
        If true, save the parameters used in the analysis in a text file.
    overwrite : bool, optional
        If true, force overwriting of existing output files.
    """

    if verbose:
        logging.getLogger().setLevel(logging.INFO)
        coloredlogs.install(level=logging.INFO)

    if loadings_df is None and model is None:
        sys.exit(
            "Please provide a loadings dataframe or a model specification."
        )

    if loadings_df is not None and threshold is None:
        sys.exit(
            "Please provide a threshold value to determine the variables to "
            "include in the CFA model."
        )

    logging.info(
        "Validating input files and creating output folder {}"
        .format(out_folder)
    )
    assert_input(in_dataset)
    assert_output_dir_exist(overwrite, out_folder, create_dir=True)

    if save_parameters:
        parameters = list(locals().items())
        with open(f"{out_folder}/parameters.txt", "w+") as f:
            for param in parameters:
                f.writelines(str(param))

    # Loading dataset.
    logging.info("Loading {}".format(in_dataset))
    if len(in_dataset) > 1:
        if id_column is None:
            sys.exit(
                "Column name for index matching is required when inputting"
                " multiple dataframes."
            )
        dict_df = {i: load_df_in_any_format(i) for i in in_dataset}
        df = merge_dataframes(dict_df, id_column)
    else:
        df = load_df_in_any_format(in_dataset[0])
    descriptive_columns = [n for n in range(0, desc_columns)]

    # Imputing missing values (or not).
    # Disabled for now, wait until we have a solution to check only columns
    # that will be used in the model.
    # if mean:
    #    logging.info("Imputing missing values using the mean method.")
    #    for column in df.columns:
    #        df[f"{column}"].fillna(df[f"{column}"].mean(), inplace=True)
    # elif median:
    #    logging.info("Imputing missing values using the median method.")
    #    for column in df.columns:
    #        df[f"{column}"].fillna(df[f"{column}"].median(), inplace=True)
    # else:
    #   logging.info(
    #        "No methods selected for imputing missing values. "
    #        "Removing them."
    #    )
    #    df.dropna(inplace=True)

    desc_col = df[df.columns[descriptive_columns]]
    df.drop(df.columns[descriptive_columns], axis=1, inplace=True)

    if loadings_df is not None:
        logging.info("Creating model's specification.")
        loadings_df = load_df_in_any_format(loadings_df)

        modeldict = {}
        for col in loadings_df.columns:
            idx = loadings_df.index[
                (loadings_df[col] >= threshold) | (loadings_df[
                    col] <= -threshold)
            ].tolist()
            modeldict[col] = idx

        mod = ""
        for key, values in modeldict.items():
            mod += f"{key} =~ {' + '.join(values)}\n"
    else:
        mod = ""
        for i in model:
            mod += f"{i}\n"

    logging.info("Performing Confirmatory Factorial Analysis (CFA) with"
                 " the following model specification:\n{}".format(mod))

    cfa_mod, scores, stats = cfa(df, mod)

    logging.info("Exporting results and statistics.")
    scores = pd.concat([desc_col, scores], axis=1)
    scores.to_excel(
        f"{out_folder}/cfa_scores.xlsx", header=True, index=False
    )

    stats.to_excel(
        f"{out_folder}/cfa_stats.xlsx", header=True, index=False
    )

    # Bootstrapping the model.
    if iterations is not None:
        logging.info("Bootstrapping the model.")
        semopy.bias_correction(cfa_mod, n=iterations)

    semopy.semplot(cfa_mod, f"{out_folder}/semplot.png",
                   plot_covs=True)
    semopy.report(cfa_mod, f"{out_folder}/CFA_report")

    # Saving the model.
    with open(f'{out_folder}/cfa_model.pkl', 'wb') as f:
        pickle.dump(cfa_mod, f)


if __name__ == "__main__":
    app()



import numpy as np
import pandas as pd
from sklearn.datasets import load_iris
import unittest

from neurostatx.statistics.utils import KNNimputation, apply_various_models
from neurostatx.utils.factor import efa, cfa
from neurostatx.utils.preprocessing import compute_pca


class TestImputation(unittest.TestCase):

    def test_impute_nans(self):
        # Define a reference dataset.
        ref_df = pd.DataFrame({'A': [1.0, 2.0, 3.0, 4.0, 5.0],
                               'B': [5.0, 4.0, 3.0, 2.0, 1.0],
                               'C': [1.0, 2.0, 3.0, 4.0, 5.0]})

        # Define a dataset to impute.
        df = pd.DataFrame({'A': [1.0, 2.0, np.nan, 4.0, 5.0],
                           'B': [5.0, 4.0, 3.0, 2.0, 1.0],
                           'C': [1.0, 2.0, 3.0, 4.0, 5.0]})

        # Impute the dataset.
        out = KNNimputation(ref_df, df, n_neighbors=1)

        # Compare the output with the expected output.
        pd.testing.assert_frame_equal(out, ref_df)

    def test_apply_efa_model(self):
        # Fetch the iris dataset.
        np.random.seed(123)
        df = pd.DataFrame(np.random.rand(100, 10))
        method = "minres"
        nfactors = 2
        rotation = 'varimax'

        # Call function
        efa_mod, ev, v, scores, loadings, communalities = \
            efa(df, method, rotation, nfactors)

        # Apply model.
        transform = apply_various_models(df, efa_mod)

        # Perform assertions
        np.testing.assert_array_equal(scores, transform)

    def test_apply_cfa_model(self):
        # Define test data
        df = pd.DataFrame(load_iris().data, columns=load_iris().feature_names)
        df.columns = ['sepal_length', 'sepal_width', 'petal_length',
                      'petal_width']
        model = ""
        model += "efa1 =~ sepal_length + petal_length\n"
        model += "efa2 =~ sepal_width + petal_length"

        # Call function
        cfa_mod, scores, stats = cfa(
            df, model)

        # Apply model.
        transform = apply_various_models(df, cfa_mod)

        # Perform assertions
        pd.testing.assert_frame_equal(scores, transform)

    def test_apply_sklearn_model(self):
        # Define test data
        X = np.random.rand(100, 5)
        n_components = 3

        # Call the function with sample data
        X_transformed, model, exp_var, components, p_value, kmo_model = \
            compute_pca(X, n_components)

        # Apply model.
        transform = apply_various_models(X, model)

        # Perform assertions
        np.testing.assert_array_equal(X_transformed, transform)


if __name__ == '__main__':
    unittest.main()



import unittest
import pandas as pd
import numpy as np

from neurostatx.utils.preprocessing import (merge_dataframes,
                                            compute_pca,
                                            remove_nans,
                                            compute_shapiro_wilk_test,
                                            compute_correlation_coefficient,
                                            rename_columns,
                                            binary_to_yes_no,
                                            get_column_indices,
                                            plot_distributions)


class TestFunctions(unittest.TestCase):

    def setUp(self):
        # Create sample data for testing
        self.df = pd.DataFrame({
            'A': [1.0, 2.0, np.nan, 4.0],
            'B': [np.nan, 2.0, 3.0, 4.0],
            'C': [1.0, 2.0, 3.0, 4.0]
        })

        self.wilk = pd.DataFrame(np.random.normal(size=3000),
                                 columns=['A'])

    def test_remove_nans(self):
        rows_with_nans, complete_rows = remove_nans(self.df)
        expected_rows_with_nans = pd.DataFrame({
            'A': [1.0, np.nan],
            'B': [np.nan, 3.0],
            'C': [1.0, 3.0]
        })
        expected_rows_complete = pd.DataFrame({
            'A': [2.0, 4.0],
            'B': [2.0, 4.0],
            'C': [2.0, 4.0]
        })

        pd.testing.assert_frame_equal(rows_with_nans.reset_index(drop=True),
                                      expected_rows_with_nans)
        pd.testing.assert_frame_equal(complete_rows.reset_index(drop=True),
                                      expected_rows_complete)

    def test_compute_shapiro_wilk_test(self):
        wilk, pvalues = compute_shapiro_wilk_test(self.wilk)
        self.assertAlmostEqual(wilk[0], 0.999, delta=0.001)

    def test_compute_correlation_coefficient(self):
        corr_mat = np.array([[1.0, 0.8, 0.2],
                             [0.8, 1.0, 0.5],
                             [0.2, 0.5, 1.0]])
        data = pd.DataFrame((
            np.random.multivariate_normal(mean=np.zeros(len(corr_mat)),
                                          cov=corr_mat,
                                          size=1000)),
                            columns=['A', 'B', 'C'])

        # Call the function with sample data
        correlation_matrix = compute_correlation_coefficient(data,
                                                             out_folder='./')

        # Check if the output is a DataFrame
        self.assertIsInstance(correlation_matrix, pd.DataFrame)

        # Check if the dimensions of the correlation matrix are correct
        n_columns = len(self.df.columns)
        self.assertEqual(correlation_matrix.shape, (n_columns, n_columns))

        # Check if the diagonal elements are all 1 (since correlation of a
        # variable with itself is 1)
        self.assertTrue(np.allclose(np.diag(correlation_matrix), 1))

        # Validate if the correlation matrix match the one used to generate the
        # sample data
        np.testing.assert_allclose(correlation_matrix.values,
                                   corr_mat, atol=0.05)

    def test_merge_dataframes(self):
        df1 = pd.DataFrame({'index': [1, 2, 4], 'A': [1, 2, 3],
                            'B': [4, 5, 6]})
        df2 = pd.DataFrame({'index': [1, 2, 4], 'C': [7, 8, 9],
                            'D': [10, 11, 12]})
        dict_df = {'df1': df1, 'df2': df2}
        merged_df = merge_dataframes(dict_df, index='index')
        expected_df = pd.DataFrame({
            'index': [1, 2, 4],
            'A': [1, 2, 3],
            'B': [4, 5, 6],
            'C': [7, 8, 9],
            'D': [10, 11, 12]
        }).set_index('index')
        pd.testing.assert_frame_equal(merged_df, expected_df)

    def test_compute_pca(self):
        # Hardly any testable assertions can be made for this function, so we
        # will only check if the output variables have the correct data types
        # and dimensions. To be updated in the future.
        X = np.random.rand(100, 5)
        n_components = 3

        # Call the function with sample data
        X_transformed, model, exp_var, components, p_value, kmo_model = \
            compute_pca(X, n_components)

        # Check if the output variables have the correct data types
        self.assertIsInstance(X_transformed, np.ndarray)
        self.assertIsInstance(exp_var, np.ndarray)
        self.assertIsInstance(components, np.ndarray)
        self.assertIsInstance(p_value, float)
        self.assertIsInstance(kmo_model, float)

        # Check if the dimensions of the transformed data are correct
        self.assertEqual(X_transformed.shape[0], X.shape[0])
        self.assertEqual(X_transformed.shape[1], n_components)

        # Check if explained variance ratio has the correct length
        self.assertEqual(len(exp_var), n_components)

        # Check if the shape of components matrix is correct
        self.assertEqual(components.shape[0], n_components)
        self.assertEqual(components.shape[1], X.shape[1])

        # Check if p-value and KMO model have valid values (not NaN)
        self.assertFalse(np.isnan(p_value))
        self.assertFalse(np.isnan(kmo_model))

    def test_rename_columns(self):
        new_df = rename_columns(self.df, ['A', 'B', 'C'], ['X', 'Y', 'Z'])

        # Assert that the columns have been renamed
        self.assertListEqual(new_df.columns.tolist(), ['X', 'Y', 'Z'])

    def test_binary_to_yes_no(self):
        df = pd.DataFrame({
            'A': [1, 0, 1, 0],
            'B': [0, 1, 0, 1]
        })

        binary_to_yes_no(df, ['A', 'B'])

        # Assert that the binary values have been converted to 'Yes' and 'No'
        self.assertListEqual(df['A'].tolist(), ['Yes', 'No', 'Yes', 'No'])
        self.assertListEqual(df['B'].tolist(), ['No', 'Yes', 'No', 'Yes'])

    def test_get_column_indices(self):
        indices = get_column_indices(self.df, ['A', 'B'])

        # Assert that the indices of columns 'A' and 'B' are correct
        self.assertListEqual(indices, [0, 1])

    def test_plot_distributions(self):
        plot_distributions(self.df, out_folder='./')


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import logging
import sys
import coloredlogs

from cyclopts import App, Parameter, Group
import pandas as pd
import networkx as nx
from typing_extensions import Annotated
from typing import List

from neurostatx.network.metrics import get_metrics_ops, get_metrics_docs
from neurostatx.io.utils import assert_input, assert_output

OPERATIONS = get_metrics_ops()

__doc__ = get_metrics_docs(OPERATIONS)

# Initializing the app.
app = App(default_parameter=Parameter(negative=()))


@app.default()
def GraphMetrics(
    operation: Annotated[
        List[str],
        Parameter(
            help=__doc__,
            show_choices=True,
            show_default=False,
            group=Group("Arguments"),
        ),
    ],
    out_file: Annotated[
        str,
        Parameter(
            show_default=False,
            group="Essential Files Options",
        ),
    ] = None,
    verbose: Annotated[
        bool,
        Parameter(
            "-v",
            "--verbose",
            group="Optional parameters",
        ),
    ] = False,
    overwrite: Annotated[
        bool,
        Parameter(
            "-f",
            "--overwrite",
            group="Optional parameters",
        ),
    ] = False,
):
    """GRAPH NETWORK METRICS
    ---------------------
    GraphMetrics is a wrapper script that allows the user to compute
    various network metrics on an existing graph network. Depending on your
    hardware and the metric you want to compute, the script could be running
    for ~10 mins for a large graph (~10 000 nodes).

    AVAILABLE METRICS
    -----------------
    Available metrics come from the Networkx implemented algorithms. As of now,
    only the algorithms that can handle undirected weighted graph are
    implemented (see operation argument). For more details regarding those
    algorithms, please see [1]. Details regarding specific operations: if NODE
    is required, you have to provide a single node label (such as 'c1'). If
    NODES is required, you can provide multiple nodes within quotation marks
    (such as "c1 c2 c3").

    REFERENCE
    ---------
    [1] [Networkx
    Algorithms](https://networkx.org/documentation/stable/reference/algorithms/index.html#)

    EXAMPLE USAGE
    -------------
    ::

        GraphMetrics --out-folder output/ eigencentrality graph.gexf membership

    Parameters
    ----------
    operation : List[str]
        Operation to perform on the graph network.
    out_file : str, optional
        Path and name of the file containing the metrics for each nodes.
        Default is : ./operation_name.xlsx
    verbose : bool, optional
        If true, produce verbose output.
    overwrite : bool, optional
        If true, force overwriting of existing output files.
    """

    if verbose:
        logging.getLogger().setLevel(logging.INFO)
        coloredlogs.install(level=logging.INFO)

    if out_file is None:
        out_file = f"{operation[0]}.xlsx"

    assert_input(operation[1])
    assert_output(overwrite, out_file, check_dir=True)

    # Validating correct input number.
    if len(operation) < 3:
        sys.exit("Incorrect number of input provided. Please see {}"
                 .format(__doc__))

    # Validating input operation exist in dict.
    if operation[0] not in OPERATIONS.keys():
        sys.exit("Operation {} not implemented.".format(operation[0]))

    # Loading graph network file.
    logging.info("Loading graph network file...")
    G = nx.read_gml(operation[1])

    # Performing operation.
    try:
        logging.info("Running {} on input network..."
                     .format(operation[0].capitalize()))
        output = OPERATIONS[operation[0]](G, *operation[2:])

    except ValueError as msg:
        logging.error("{} operation failed.".format(operation[0].capitalize()))
        logging.error(msg)
        return

    # Exporting results in an .xlsx file.
    logging.info("Exporting results here: {}".format(out_file))
    output = pd.DataFrame.from_dict([output]).T
    output.index.name = "nodes"
    output.columns = [operation[0]]
    output.to_excel(out_file, header=True, index=True)


if __name__ == "__main__":
    app()



from factor_analyzer.factor_analyzer import (calculate_kmo,
                                             calculate_bartlett_sphericity)
import matplotlib.pyplot as plt
from scipy.stats import shapiro
from sklearn.decomposition import PCA
import seaborn as sns
from tqdm import tqdm


def remove_nans(df):
    """
    Clean up dataset by removing all rows containing NaNs.
    :param df:      Pandas dataframe.
    :return:        One df containing the removed rows and one df with the
                    complete rows.
    """
    rows_with_nans = df[df.isna().any(axis=1)]
    complete_rows = df.drop(index=rows_with_nans.index)

    return rows_with_nans, complete_rows


def rename_columns(df, old_names, new_names):
    """
    Function renaming specific columns according to a list of new and old
    column names.

    :param df:              Pandas dataframe object.
    :param old_names:       List of old column name as strings.
    :param new_names:       List of new column name as strings.
    :return:
    Pandas dataframe object containing the renamed columns.
    """
    if len(old_names) != len(new_names):
        raise ValueError("Number of old names and new names must be the same.")

    for i, old_name in enumerate(old_names):
        if old_name not in df.columns:
            raise ValueError(f"Column {old_name} not found in DataFrame.")

    new_names_set = set(new_names)
    if len(new_names_set) != len(new_names):
        raise ValueError("New names contain duplicates.")

    new_df = df.copy()
    for i in range(len(old_names)):
        if old_names[i] in new_df.columns:
            new_df.rename(columns={old_names[i]: new_names[i]}, inplace=True)
    return new_df


def binary_to_yes_no(df, cols):
    """
    Function to change binary answers (1/0) to Yes or No in specific columns
    from a Pandas Dataframe.
    *** Please validate that yes and no are assigned to the correct values,
    default behavior is yes = 1 and no = 0 ***
    :param df:              Pandas' dataframe object.
    :param cols:            List of column names.
    :return:
    Pandas' dataframe object with changed binary answers.
    """
    for col in cols:
        if df[col].isin([0.0, 1.0, 2.0, "nan"]).any():
            df[col] = df[col].apply(
                lambda x: "Yes"
                if x == 1
                else "No"
                if x == 0
                else "Don't know or missing value"
            )
    return df


def get_column_indices(df, column_names):
    """
    Function to extract column index based on a list of column names.
    :param df:              Pandas dataframe object.
    :param column_names:    List of column names as strings.
    :return:
    List of column index.
    """
    indices = []
    for name in column_names:
        try:
            index = df.columns.get_loc(name)
            indices.append(index)
        except KeyError:
            print(f"Column '{name}' not found in DataFrame.")
    return indices


def plot_distributions(df, out_folder, context="poster", font_scale=1):
    """
    Script to visualize distribution plots for a complete dataframe.
    :param df:              Pandas dataframe.
    :param out_folder:      Path to the output folder.
    :param context:         Style to apply to the plots.
    :param font_scale:      Font Scale
    :return:
    """

    # Setting plotting parameters.
    plt.rcParams["figure.figsize"] = [10, 7]
    plt.rcParams["figure.autolayout"] = True
    sns.set_style("white")
    sns.set_context(f"{context}", font_scale)

    # Iterating over columns for plotting variables.
    for var in tqdm(df.columns):
        f, axes = plt.subplots(1, 2)
        sns.histplot(data=df, x=var, fill=True, kde=True, ax=axes[0])
        sns.ecdfplot(data=df, x=var, ax=axes[1])
        plt.savefig(f"{out_folder}/{var}.png")
        plt.close()


def compute_shapiro_wilk_test(df):
    """
    Function computing the normality statistic using the Shapiro Wilk's test
    for normality and outputting W and p values.
    :param df:      Pandas dataframe.
    :return:        Shapiro-Wilk values (W) and associated p-values.
    """

    wilk = []
    pvalues = []

    for var in tqdm(df.columns):
        var_data = df[var].values
        w, pval = shapiro(var_data)
        wilk.append(w)
        pvalues.append(pval)

    return wilk, pvalues


def compute_correlation_coefficient(
    df, out_folder, context="poster", font_scale=0.2, cmap=None, annot=False
):
    """
    Function to compute a correlation matrix for all variables in a dataframe.
    :param df:              Pandas dataframe.
    :param out_folder:      Path to the output folder.
    :param context:         Style to apply to the plots.
    :param font_scale:      Font scale.
    :param cmap:            Cmap to use in the heatmap.
    :param annot:           Flag to write correlation values inside the
                            heatmap squares.
    :return:
    Correlation matrix with pearson correlation coefficients.
    """

    # Setting plotting parameters.
    plt.rcParams["figure.figsize"] = [20, 15]
    plt.rcParams["figure.autolayout"] = True
    sns.set_style("white")
    sns.set_context(f"{context}", font_scale)

    corr_mat = df.corr()
    sns.heatmap(
        corr_mat,
        cmap=cmap,
        annot=annot,
        square=True,
        xticklabels=True,
        yticklabels=True,
        cbar=True,
    )
    plt.savefig(f"{out_folder}/correlation_heatmap.png")

    return corr_mat


def merge_dataframes(dict_df, index, repeated_columns=False):
    """
    Function to merge a variable number of dataframe by matching the values of
    a specific column (hereby labeled as index.) Index values must appear only
    once in the dataframe for the function to work.
    :param dict_df:             Dictionary of pandas dataframe.
    :param index:               String of the name of the column to use as
                                index (needs to be the same across all
                                dataframes).
    :param repeated_columns     Flag to use if column name are repeated across
                                dataframe to merge.
    :return:                    Joint large pandas dataframe.
    """

    keys = list(dict_df.keys())
    for k in keys:
        dict_df[k] = dict_df[k].set_index(f"{index}")

    if repeated_columns:
        out = dict_df[keys[0]]
        for k in keys[1: len(keys)]:
            out = out.join(dict_df[k], lsuffix="a", rsuffix="b")
    else:
        out = dict_df[keys[0]].join([dict_df[k] for k in keys[1: len(keys)]])

    return out


def compute_pca(X, n_components):
    """
    Function compute PCA decomposition on a dataset.

    Args:
        X (Array):                  Data array.
        n_components (int):         Number of components.

    Return:
        X (Array):                  Transformed data array.
        exp_var (Array):            Explained variance.
        components (Array):         Components.
        p_value (float):            Bartlett's p-value.
        kmo_model (float):          KMO model.

    """
    chi_square_value, p_value = calculate_bartlett_sphericity(X)
    kmo_all, kmo_model = calculate_kmo(X)
    pca = PCA(n_components=n_components).fit(X)
    X = pca.transform(X)
    exp_var = pca.explained_variance_ratio_
    components = pca.components_

    return X, pca, exp_var, components, p_value, kmo_model



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import coloredlogs
import logging
import os
import sys
from joblib import load

from cyclopts import App, Parameter
import numpy as np
import pandas as pd
from skfuzzy import cmeans_predict
from typing import List
from typing_extensions import Annotated

from neurostatx.io.utils import (load_df_in_any_format, assert_input,
                                 assert_output_dir_exist)
from neurostatx.clustering.viz import (
    plot_parallel_plot,
    radar_plot)
from neurostatx.io.viz import flexible_barplot
from neurostatx.utils.preprocessing import merge_dataframes, compute_pca
from neurostatx.clustering.distance import DistanceMetrics


# Initializing the app.
app = App(default_parameter=Parameter(negative=()))


@app.default()
def PredictFuzzyMembership(
    in_dataset: Annotated[
        List[str],
        Parameter(
            show_default=False,
            group="Essential Files Options",
        ),
    ],
    in_cntr: Annotated[
        str,
        Parameter(
            show_default=False,
            group="Essential Files Options",
        ),
    ],
    id_column: Annotated[
        str,
        Parameter(
            show_default=False,
            group="Essential Files Options",
        ),
    ],
    desc_columns: Annotated[
        int,
        Parameter(
            show_default=False,
            group="Essential Files Options",
        ),
    ],
    out_folder: Annotated[
        str,
        Parameter(
            show_default=False,
            group="Essential Files Options",
        ),
    ] = "./output/",
    m: Annotated[
        float,
        Parameter(
            show_default=True,
            group="Clustering Options",
        ),
    ] = 2,
    error: Annotated[
        float,
        Parameter(
            show_default=True,
            group="Clustering Options",
        ),
    ] = 1e-6,
    maxiter: Annotated[
        int,
        Parameter(
            show_default=True,
            group="Clustering Options",
        ),
    ] = 1000,
    metric: Annotated[
        DistanceMetrics,
        Parameter(
            show_default=True,
            group="Clustering Options",
        ),
    ] = DistanceMetrics.euclidean,
    pca: Annotated[
        bool,
        Parameter(
            "--pca",
            show_default=True,
            group="Clustering Options",
        ),
    ] = False,
    pca_model: Annotated[
        str,
        Parameter(
            show_default=True,
            group="Clustering Options",
        ),
    ] = None,
    parallelplot: Annotated[
        bool,
        Parameter(
            show_default=True,
            group="Visualization Options",
        ),
    ] = False,
    radarplot: Annotated[
        bool,
        Parameter(
            show_default=True,
            group="Visualization Options",
        ),
    ] = True,
    cmap: Annotated[
        str,
        Parameter(
            show_default=True,
            group="Visualization Options",
        ),
    ] = "magma",
    verbose: Annotated[
        bool,
        Parameter(
            "-v",
            "--verbose",
            group="Optional parameters",
        ),
    ] = False,
    save_parameters: Annotated[
        bool,
        Parameter(
            "-s",
            "--save_parameters",
            group="Optional parameters",
        ),
    ] = False,
    overwrite: Annotated[
        bool,
        Parameter(
            "-f",
            "--overwrite",
            group="Optional parameters",
        ),
    ] = False,
):
    """FUZZY MEMBERSHIP PREDICTION
    ---------------------------
    This script will predict the membership matrix of a dataset using a
    trained Cmeans model (only the centroids are necessary for the prediction).

    PARAMETERS
    ----------
    Details regarding the parameters can be seen below. Regarding the
    --m parameter, it defines the degree of fuzziness of the resulting
    membership matrix. Using --m 1 will returns crisp clusters, whereas
    --m >1 will returned more and more fuzzy clusters. It is recommended
    to use the same m value as used during training.

    EXAMPLE USAGE
    -------------
    ::

        PredictFuzzyMembership --in_dataset dataset.xlsx --in_cntr
        centroids.xlsx --id_column ID --desc_columns 1 --out_folder
        predicted_membership_matrix/ --m 2 --error 1e-6 --maxiter 1000
        --metric euclidean --verbose --save_parameters --overwrite

    Parameters
    ----------
    in_dataset : List[str]
        Input dataset(s) to filter. If multiple files are provided as input,
        will be merged according to the subject id columns.
    in_cntr : str
        Centroid file to use for prediction. Should come from a trained Cmeans
        model (such as ``FuzzyClustering``).
    id_column : str
        Name of the column containing the subject's ID tag. Required for proper
        handling of IDs and merging multiple datasets.
    desc_columns : int
        Number of descriptive columns at the beginning of the dataset to
        exclude in statistics and descriptive tables.
    out_folder : str, optional
        Output folder for the predicted membership matrix.
    m : float, optional
        Exponentiation value to apply on the membership function, will
        determined the degree of fuzziness of the membership matrix.
    error : float, optional
        Error threshold for convergence stopping criterion.
    maxiter : int, optional
        Maximum number of iterations to perform.
    metric : DistanceMetrics, optional
        Metric to use to compute distance between original points and clusters
        centroids.
    pca : bool, optional
        If set, will perform PCA decomposition to 3 components before
        clustering.
    pca_model : str, optional
        If set, will load a pre-trained PCA model to apply on the dataset.
    parallelplot : bool, optional
        If true, will output parallel plot for each cluster solution. Default
        is False.
    radarplot : bool, optional
        If true, will output radar plot for each cluster solution. Default is
        True.
    cmap : str, optional
        Colormap to use for plotting. Default is "magma". See
        [Matplotlib](https://matplotlib.org/stable/tutorials/colors/colormaps.html).
    verbose : bool, optional
        If true, produce verbose output.
    save_parameters : bool, optional
        If true, will save input parameters to .txt file.
    overwrite : bool, optional
        If true, force overwriting of existing output files.
    """

    if verbose:
        logging.getLogger().setLevel(logging.INFO)
        coloredlogs.install(level=logging.INFO)

    assert_input(in_dataset)
    assert_output_dir_exist(overwrite, out_folder, create_dir=True)

    if save_parameters:
        parameters = list(locals().items())
        with open(f"{out_folder}/parameters.txt", "w+") as f:
            for param in parameters:
                f.writelines(str(param))

    # Loading dataframe.
    logging.info("Loading dataset(s)...")
    if len(in_dataset) > 1:
        if id_column is None:
            sys.exit(
                "Column name for index matching is required when inputting "
                "multiple dataframe."
            )
        dict_df = {i: load_df_in_any_format(i) for i in in_dataset}
        raw_df = merge_dataframes(dict_df, id_column)
    else:
        raw_df = load_df_in_any_format(in_dataset[0])
    descriptive_columns = [n for n in range(0, desc_columns)]

    cntr = load_df_in_any_format(in_cntr)
    cntr.drop(cntr.columns[0], axis=1, inplace=True)

    # Creating the array.
    desc_data = raw_df[raw_df.columns[descriptive_columns]]
    df_for_clust = raw_df.drop(
        raw_df.columns[descriptive_columns], axis=1, inplace=False
    ).astype("float")
    X = df_for_clust.values

    # Decomposing into 2 components if asked.
    if pca:
        if pca_model is not None:
            logging.info("Loading PCA model...")
            pca_model = load(pca_model)
            X = pca_model.transform(X)
        else:
            logging.info("Applying PCA dimensionality reduction.")
            X, variance, components, chi, kmo = compute_pca(X, 3)
            logging.info(
                "Bartlett's test of sphericity returned a p-value of {} and "
                "Keiser-Meyer-Olkin (KMO)"
                " test returned a value of {}.".format(chi, kmo)
            )
            # Exporting variance explained data.
            os.mkdir(f"{out_folder}/PCA/")
            var_exp = pd.DataFrame(variance, columns=["Variance Explained"])
            var_exp.to_excel(
                f"{out_folder}/PCA/variance_explained.xlsx", index=True,
                header=True
            )
            components_df = pd.DataFrame(components,
                                         columns=df_for_clust.columns)
            components_df.to_excel(f"{out_folder}/PCA/components.xlsx",
                                   index=True,
                                   header=True)
            out = pd.DataFrame(X, columns=["Component #1", "Component #2",
                                           "Component #3"])
            out.to_excel(f"{out_folder}/PCA/transformed_data.xlsx", index=True,
                         header=True)

            flexible_barplot(
                components,
                df_for_clust.columns,
                3,
                title="Loadings values for the two components.",
                filename=f"{out_folder}/PCA/barplot_loadings.png",
                ylabel="Loading value")

    logging.info("Predicting membership matrix...")
    u, u0, d, jm, p, fpc = cmeans_predict(
        X.T,
        cntr.values,
        m=m,
        error=error,
        maxiter=maxiter,
        metric=metric,
        init=None,
        seed=42,
    )

    # Saving results.
    logging.info("Saving results...")
    member = pd.DataFrame(
        u.T,
        index=None,
        columns=[f"Cluster #{n+1}" for n in range(u.shape[0])],
    )
    out = pd.concat([raw_df, member], axis=1)
    out.to_excel(
        f"{out_folder}/predicted_membership_matrix.xlsx",
        header=True,
        index=False,
    )

    os.mkdir(f"{out_folder}/PARALLEL_PLOTS/")
    os.mkdir(f"{out_folder}/BARPLOTS/")
    os.mkdir(f"{out_folder}/RADAR_PLOTS/")
    membership = np.argmax(u, axis=0)
    if parallelplot:
        plot_parallel_plot(
            df_for_clust,
            membership,
            mean_values=True,
            output=f"{out_folder}/PARALLEL_PLOTS/parallel_plot_{u.shape[0]}"
                   "clusters.png",
            title=f"Parallel Coordinates plot for {u.shape[0]} clusters "
            "solution.",
            cmap=cmap
        )
    if radarplot:
        radar_plot(
            df_for_clust,
            membership,
            title=f"Radar plot of {u.shape[0]} clusters solution.",
            frame='circle',
            cmap=cmap,
            output=(
                f"{out_folder}/RADAR_PLOTS/radar_plot_{u.shape[0]}clusters.png"
            ),
        )


if __name__ == "__main__":
    app()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import coloredlogs
import logging

import matplotlib.pyplot as plt
import pandas as pd
from cyclopts import App, Parameter
from typing import List
from typing_extensions import Annotated
import seaborn as sns

from neurostatx.io.utils import (assert_input, assert_output_dir_exist,
                                 load_df_in_any_format)
from neurostatx.clustering.metrics import compute_rand_index

# Initializing the app.
app = App(default_parameter=Parameter(negative=()))


@app.default()
def CompareClustering(
    in_dataset: Annotated[
        List[str],
        Parameter(
            show_default=False,
            group="Essential Files Options",
        ),
    ],
    id_column: Annotated[
        str,
        Parameter(
            show_default=False,
            group="Essential Files Options",
        ),
    ],
    desc_columns: Annotated[
        int,
        Parameter(
            show_default=False,
            group="Essential Files Options",
        ),
    ],
    out_folder: Annotated[
        str,
        Parameter(
            show_default=False,
            group="Essential Files Options",
        ),
    ],
    columns_name: Annotated[
        List[str],
        Parameter(
            group="Visualization Options",
            show_default=True,
        )
    ] = [],
    cmap: Annotated[
        str,
        Parameter(
            show_default=True,
            group="Visualization Options",
        )
    ] = "magma",
    title: Annotated[
        str,
        Parameter(
            show_default=True,
            group="Visualization Options",
        )
    ] = 'Adjusted Rand Index Heatmap',
    verbose: Annotated[
        bool,
        Parameter(
            "-v",
            "--verbose",
            group="Optional parameters",
        ),
    ] = False,
    save_parameters: Annotated[
        bool,
        Parameter(
            "-s",
            "--save_parameters",
            group="Optional parameters",
        ),
    ] = False,
    overwrite: Annotated[
        bool,
        Parameter(
            "-f",
            "--overwrite",
            group="Optional parameters",
        ),
    ] = False,
):
    """COMPARE CLUSTERING RESULTS
    --------------------------
    CompareClustering is a script that compares clustering results
    from multiple solutions using the Adjusted Rand Index (ARI) and produces a
    heatmap of the results.

    ADJUSTED RAND INDEX
    -------------------
    The Adjusted Rand Index (ARI) is a measure of similarity between two
    clustering results. It relies on comparing the predicted labels to the
    ground truth labels (i.e. the true clustering). The ARI is a number
    between -1 and 1, where 1 means that the two clustering results are
    identical, 0 means that the two clustering results are independent
    (as good as random labelling) and -1 means that the two clustering
    results are completely different. The ARI is an extension of the Rand
    Index (RI) that takes into account the fact that the RI is expected to
    be higher for large number of clusters.

    REFERENCES
    ----------
    [1] [Hubert, L., & Arabie, P. (1985). Comparing partitions. Journal of
    classification, 2(1), 193-218.](https://doi.org/10.1007/BF01908075)

    [2] [Rand, W. M. (1971). Objective criteria for the evaluation of
    clustering methods. Journal of the American Statistical Association,
    66(336), 846-850.](https://doi.org/10.2307/2284239)

    [3] [D. Steinley, Properties of the Hubert-Arabie adjusted Rand index,
    Psychological Methods
    2004](https://psycnet.apa.org/doi/10.1037/1082-989X.9.3.386)

    [4] [Scikit-Learn Documentation - Adjusted Rand Index
    (ARI)](https://scikit-learn.org/stable/modules/generated/sklearn.metrics.adjusted_rand_score.html)

    EXAMPLE USAGE
    -------------
    ::

        CompareClustering --in_dataset dataset1.csv --in_dataset dataset2.csv
        --in_dataset dataset3.csv --id_column ID --desc_columns 1 --out_folder
        ./ --columns_name dataset1 dataset2 dataset3 --title "ARI Heatmap"
        --verbose

    Parameters
    ----------
    in_dataset : List[str]
        Input dataset(s) (at least 2 are expected to produce a comparison).
    id_column : str
        Name of the column containing the subject's ID tag. Required for
        proper handling of IDs and merging multiple datasets.
    desc_columns : int
        Number of descriptive columns at the beginning of the dataset to
        exclude in statistics and descriptive tables.
    out_folder : str
        Output folder containing the results.
    columns_name : List[str], optional
        Name given to each input dataset (needs to be in the same order as
        the input datasets).
    cmap : str, optional
        Name of the colormap to use. Defaults to "magma". See
        https://matplotlib.org/stable/tutorials/colors/colormaps.html
    title : str, optional
        Heatmap title.
    verbose : bool, optional
        If true, produce verbose output.
    save_parameters : bool, optional
        If true, will save input parameters to .txt file.
    overwrite : bool, optional
        If true, force overwriting of existing output files.
    """

    if verbose:
        logging.getLogger().setLevel(logging.INFO)
        coloredlogs.install(level=logging.INFO)

    assert_input(in_dataset)
    assert_output_dir_exist(overwrite, out_folder, create_dir=True)

    if save_parameters:
        parameters = list(locals().items())
        with open(f"{out_folder}/parameters.txt", "w+") as f:
            for param in parameters:
                f.writelines(str(param))

    logging.info("Loading datasets...")
    # Loading all datasets into a dictionary.
    assert len(in_dataset) >= 2, "At least 2 datasets are required for "
    "                            comparison."
    dict_df = {i: load_df_in_any_format(df) for i, df in enumerate(in_dataset)}

    # Dropping desc column.
    descriptive_columns = [n for n in range(0, desc_columns)]
    for k in dict_df.keys():
        dict_df[k] = dict_df[k].drop(
            dict_df[k].columns[descriptive_columns], axis=1)

    logging.info("Computing Adjusted Rand Index...")
    # Comparison between each clustering results.
    ari = compute_rand_index(dict_df)

    logging.info("Plotting and saving results...")
    # Exporting symmetric matrix.
    if len(columns_name) == 0:
        columns_name = [f'{i+1}' for i in range(0, len(in_dataset))]
    mat = pd.DataFrame(ari, columns=columns_name,
                       index=columns_name)
    mat.to_excel(f'{out_folder}/ari_matrix.xlsx', index=True, header=True)

    # Create heatmap.
    with plt.rc_context(
        {"font.family": "Sans Serif", "font.size": 12, "font.weight": "normal",
         "axes.titleweight": "bold"}
    ):
        fig, ax = plt.subplots(figsize=(12, 10))
        sns.heatmap(mat, annot=True, ax=ax, cmap=cmap, linewidths=1,
                    vmin=-1, vmax=1)
        ax.set_title(title)
        plt.savefig(f"{out_folder}/ari_heatmap.png")
        plt.close()


if __name__ == "__main__":
    app()



# !/usr/bin/env python3
# -*- coding: utf-8 -*-

import os
import tempfile

from neurostatx.io.download import get_home, download_data, save_files_dict


download_data(save_files_dict(), keys=["data.zip"])
tmp_dir = tempfile.TemporaryDirectory()


def test_help(script_runner):
    ret = script_runner.run(["CompareGraphs", "-h"])

    assert ret.success


def test_compute_graph_network(script_runner):
    os.chdir(os.path.expanduser(tmp_dir.name))
    in_graph = os.path.join(get_home(), "data/graph_file.gml")
    out_folder = os.path.join(get_home(), "data/compare_graph/")

    ret = script_runner.run([
        "CompareGraphs",
        "--in-graph1", in_graph,
        "--weight", "membership",
        "--percentile", 80,
        "--in-graph2", in_graph,
        "--out-folder", out_folder,
        "-f"]
    )

    assert ret.success



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import coloredlogs
import logging
import os
import warnings

from cyclopts import App, Parameter
import matplotlib.pyplot as plt
import networkx as nx
import numpy as np
import pandas as pd
from sklearn.preprocessing import scale
from sklearn.linear_model import LogisticRegressionCV
from sklearn.model_selection import train_test_split
from sklearn.preprocessing import LabelBinarizer
from sklearn.metrics import RocCurveDisplay
from typing import List
from typing_extensions import Annotated

from neurostatx.io.utils import (assert_input, assert_output_dir_exist)
from neurostatx.network.utils import fetch_attributes_df, fetch_edge_data
from neurostatx.statistics.models import (permutation_testing, ScoringMethod,
                                          Penalty, Solver)
from neurostatx.io.viz import generate_coef_plot, flexible_hist

# Initializing the app.
app = App(default_parameter=Parameter(negative=()))


@app.default()
def LogisticRegression(
    in_graph: Annotated[
        str,
        Parameter(
            show_default=False,
            group="Essential Files Options",
        ),
    ],
    out_folder: Annotated[
        str,
        Parameter(
            show_default=False,
            group="Essential Files Options",
        ),
    ],
    attributes: Annotated[
        List[str],
        Parameter(
            group="Model Parameters",
            show_default=False,
        ),
    ],
    covariates: Annotated[
        List[str],
        Parameter(
            group="Model Parameters",
            show_default=True,
        ),
    ] = None,
    weight: Annotated[
        str,
        Parameter(
            group="Model Parameters",
            show_default=True,
        ),
    ] = "membership",
    splits: Annotated[
        int,
        Parameter(
            group="Model Parameters",
            show_default=True,
        ),
    ] = 10,
    test_size: Annotated[
        float,
        Parameter(
            group="Model Parameters",
            show_default=True,
        ),
    ] = 0.2,
    cs: Annotated[
        int,
        Parameter(
            group="Model Parameters",
            show_default=True,
        ),
    ] = 10,
    max_iter: Annotated[
        int,
        Parameter(
            group="Model Parameters",
            show_default=True,
        ),
    ] = 1000,
    penalty: Annotated[
        Penalty,
        Parameter(
            group="Model Parameters",
            show_default=True,
            show_choices=True,
        ),
    ] = Penalty.l2,
    solver: Annotated[
        Solver,
        Parameter(
            group="Model Parameters",
            show_default=True,
            show_choices=True,
        ),
    ] = Solver.lbfgs,
    permutations: Annotated[
        int,
        Parameter(
            group="Model Parameters",
            show_default=True,
        ),
    ] = 1000,
    scoring: Annotated[
        ScoringMethod,
        Parameter(
            group="Model Parameters",
            show_default=True,
        ),
    ] = ScoringMethod.r2,
    processes: Annotated[
        int,
        Parameter(
            group="Model Parameters",
            show_default=True,
        ),
    ] = 1,
    plot_distributions: Annotated[
        bool,
        Parameter(
            show_default=True,
            group="Visualization Options",
        ),
    ] = False,
    verbose: Annotated[
        bool,
        Parameter(
            "-v",
            "--verbose",
            group="Optional parameters",
        ),
    ] = False,
    save_parameters: Annotated[
        bool,
        Parameter(
            "-s",
            "--save_parameters",
            group="Optional parameters",
        ),
    ] = False,
    overwrite: Annotated[
        bool,
        Parameter(
            "-f",
            "--overwrite",
            group="Optional parameters",
        ),
    ] = False
):
    """LOGISTIC REGRESSION ANALYSIS
    --------------------------------------
    This script performs a Logistic Regression on a graph
    using the edges' weights as predictors and the nodes' attributes as
    response variable. The script will perform a cross-validation within a
    training dataset before testing the model on test data. It will then
    perform a permutation testing to determine if the model is statistically
    significant. Finally, it will output the coefficients and statistics
    as well as plots of the distributions of the attributes and edges' weights
    and the coefficients.

    PREPROCESSING
    -------------
    The script will scale the data to unit variance and zero mean and will
    perform a log transformation on the edges' weights (for now, it assumes
    that the weights represent a membership value resulting from a fuzzy
    clustering analysis).

    NODES' ATTRIBUTES
    -----------------
    The script takes only one graph file as input. The graph file must be in
    .gml format. The script will then fetch the attributes from the graph
    file and will perform the analysis on the attributes and edges'
    weights. At least one attribute needs to be provided in order to fit a
    model. To set attributes to the nodes in the graph file,
    please see AddNodesAttributes.

    SCORING OPTIONS
    ---------------
    The script will perform a permutation testing to determine if the model is
    statistically significant. The script will compute the p-value for the
    permutation testing using the area under the curve (AUC) by default.
    However, the user can also choose multiple scores to compute the p-value.
    The available scores can be seen in [1]. The equation used to compute
    the single-tailed p-value is:

        p-value = ∑(score_perm >= score) / (nb_permutations)

    COEFFICIENT SIGNIFICANCE
    ------------------------
    The script will also compute the p-value for the coefficients using the
    permutation testing. The p-value for the coefficients is computed by
    comparing the coefficients obtained from the model with the
    coefficients obtained from the permutation testing. The equation used to
    compute the two-tailed p-value is:

        p-value = ∑(abs(coef_perm) >= abs(coef)) / (nb_permutations)

    REFERENCES
    ----------
    [1]
    https://scikit-learn.org/stable/modules/model_evaluation.html#scoring-parameter

    EXAMPLE USAGE
    -------------
    ::

        LogisticRegression --in-graph graph.gexf --out-folder output_folder -v
        -s

    Parameters
    ----------
    in_graph : str
        Graph file containing the data for the model.
    out_folder : str
        Output folder.
    attributes : List[str],
        Attributes names to include in the model. Must be present in the graph
        file. At least one attribute is required.
    covariates : List[str], optional
        Covariates to include in the model. Must be present in the graph file.
    weight : str, optional
        Edge weight to use for the model.
    splits : int, optional
        Number of splits to use for the cross-validation.
    test_size : float, optional
        Size of the testing set. Must be between 0 and 1.
    cs : int, optional
        Inverse of regularization strength. Smaller values specify stronger
        regularization.
    max_iter : int, optional
        Maximum number of iterations for the solver.
    penalty : Penalty, optional
        Regularization penalty to use for the LogisticRegression model.
    solver : Solver, optional
        Solver to use for the LogisticRegression model.
    permutations : int, optional
        Number of permutations to use for the permutation testing.
    scoring : ScoringMethod, optional
        Scoring method to use for the permutation testing.
    processes : int, optional
        Number of processes to use for the cross-validation.
    verbose : bool, optional
        If true, produce verbose output.
    plot_distributions : bool, optional
        If true, will plot the distributions of the edges' weights.
    save_parameters : bool, optional
        If true, will save input parameters to .txt file.
    overwrite : bool, optional
        If true, force overwriting of existing output files.
    """

    warnings.filterwarnings("ignore")

    if verbose:
        logging.getLogger().setLevel(logging.INFO)
        coloredlogs.install(level=logging.INFO)

    assert_input(in_graph)
    assert_output_dir_exist(overwrite, out_folder, create_dir=True)

    # Creating output structure.
    os.makedirs(f"{out_folder}/Distributions", exist_ok=True)
    os.makedirs(f"{out_folder}/Coefficients", exist_ok=True)

    # Saving parameters
    if save_parameters:
        parameters = list(locals().items())
        with open(f"{out_folder}/nodes_attributes_parameters.txt", "w+") as f:
            for param in parameters:
                f.writelines(str(param))

    # Load graph file.
    logging.info("Loading graph and dataset.")
    G = nx.read_gml(in_graph)

    # Fetching covariates, attributes and edge data.
    logging.info("Fetching covariates, attributes and edge data.")
    attributes_df = fetch_attributes_df(G, attributes)
    edge_df = fetch_edge_data(G, weight)
    covariates_df = fetch_attributes_df(G, covariates)

    # Preprocessing data.
    logging.info("Scaling data.")
    edge = edge_df.apply(lambda x: np.log(x))
    predictor = pd.concat([pd.DataFrame(scale(edge), columns=edge_df.columns,
                                        index=edge_df.index),
                          covariates_df], axis=1)

    # Performing Cross-Validation.
    logging.info("Fitting model with permutation testing.")

    # Splitting into a training and testing set.
    X_train, X_test, y_train, y_test = train_test_split(
        predictor,
        attributes_df,
        test_size=test_size,
        random_state=1234
    )

    if penalty == Penalty.elasticnet:
        l1_ratios = [0.1, 0.5, 0.7, 0.9]
    else:
        l1_ratios = None

    # Stratified KFold.
    # kf_10 = StratifiedKFold(n_splits=splits, shuffle=True, random_state=1)

    classifier = LogisticRegressionCV(
        Cs=cs,
        cv=splits,
        max_iter=max_iter,
        penalty=penalty,
        scoring=scoring,
        solver=solver,
        l1_ratios=l1_ratios,
        n_jobs=processes
    )

    mod, score, coef, perm_score, score_pval, coef_perm, coef_pval = \
        permutation_testing(
            classifier,
            X_train,
            y_train.values.ravel(),
            splits=splits,
            nb_permutations=permutations,
            scoring=scoring,
            processes=processes,
            verbose=verbose
        )

    logging.info("Exporting statistics and plots.")
    for i in range(0, len(attributes_df.columns)):
        display = RocCurveDisplay.from_predictions(
            LabelBinarizer().fit(y_train).transform(y_test)[:, i],
            mod.predict_proba(X_test)[:, 1],
            name="Logistic Regression",
            color='black',
            plot_chance_level=True
        )
        # Set labels and title.
        display.ax_.set(
            xlabel="False Positive Rate (FPR)",
            ylabel="True Positive Rate (TPR)",
            title=f"Receiver Operating Characteristic (ROC) Curve for \
            {attributes_df.columns[i]}"
        )
        plt.tight_layout()
        plt.savefig(f"{out_folder}/roc_curve_{attributes_df.columns[i]}.png")
        plt.close()

    # Generating distribution plots.
    if plot_distributions:
        edge_viz = pd.DataFrame(edge, columns=edge_df.columns)
        flexible_hist(edge_viz,
                      f'{out_folder}/Distributions/edges_distributions.png',
                      cmap="magma", title="Edges' Distributions",
                      xlabel="Edges' Weights", ylabel="Density")

    # Generating coefficient plot.
    coef = {
        f'coef{i+1}': coef[:, i].T for i in range(0, coef.shape[1])
    }
    coef['varname'] = predictor.columns
    coef_df = pd.DataFrame(coef)
    coef_df.to_excel(f"{out_folder}/Coefficients/coefficients.xlsx")

    coef_pval_df = pd.DataFrame(coef_pval,
                                index=predictor.columns,
                                columns=attributes_df.columns)
    coef_pval_df.to_excel(f"{out_folder}/Coefficients/coefficients_pval.xlsx")

    stats = pd.DataFrame([score, np.median(perm_score), score_pval],
                         columns=['Statistics'],
                         index=['Score', 'Median Permuted Score', 'P-Value'])
    stats.to_excel(f"{out_folder}/statistics.xlsx", header=True, index=True)

    for i in range(0, len(attributes_df.columns)):
        generate_coef_plot(
            coef_df,
            coef_pval[:, i],
            coefname=f'coef{i+1}',
            varname='varname',
            output=f"{out_folder}/Coefficients/coef_plot_cluster_{i+1}.png"
        )


if __name__ == "__main__":
    app()



import unittest
from sklearn import datasets
from sklearn.cross_decomposition import PLSRegression
from sklearn.model_selection import train_test_split, permutation_test_score
from sklearn.metrics import mean_squared_error

from neurostatx.statistics.models import plsr_cv, permutation_testing


class TestPLSRFunctions(unittest.TestCase):

    def setUp(self) -> None:
        # Generate synthetic data
        # Not optimal, but gives a certain quality control over the results.

        # Load the Diabetes dataset
        diabetes = datasets.load_diabetes()

        # Extract features (X) and target variable (y)
        self.X = diabetes.data
        self.y = diabetes.target

        # Split the data into training and testing sets
        X_train, X_test, y_train, y_test = train_test_split(self.X, self.y,
                                                            test_size=0.2,
                                                            random_state=1234)

        # Initialize PLS model with the desired number of components
        n_components = 3
        self.pls_model = PLSRegression(n_components=n_components)

        # Fit the model on the training data
        self.pls_model.fit(X_train, y_train)

        # Predictions on the test set
        y_pred = self.pls_model.predict(X_test)

        # Evaluate the model performance
        self.r_squared = self.pls_model.score(X_test, y_test)
        self.mse_expected = mean_squared_error(y_test, y_pred)

        return super().setUp()

    def test_plsr_cv(self):
        # Using the pls_cv function.
        plsr, mse, score_c, score_cv, rscore, mse_c, mse_cv = plsr_cv(
            self.X, self.y, nb_comp=3, splits=10, processes=1)

        # Validating score and mse are within a close range.
        self.assertAlmostEqual(score_c, self.r_squared, delta=0.05)
        self.assertAlmostEqual(mse_c, self.mse_expected, delta=100)

    def test_permutation_testing(self):
        # Since this implementation is a modified version from
        # sklearn's permutation testing, we will assert the p-value obtained
        # from the permutation testing is identical.
        # Supply the precomputed PLSR model to the permutation testing function
        mod, sc, c, psc, spval, pc, cpval = permutation_testing(
            self.pls_model, self.X, self.y, nb_permutations=100, splits=10,
            processes=1)

        # Using the permutation testing function from sklearn
        sk_sc, sk_psc, sk_pval = permutation_test_score(
            self.pls_model, self.X, self.y, n_permutations=100,
            cv=10, n_jobs=1)

        # Assert the score and p-values are identical.
        self.assertEqual(sc, sk_sc)
        self.assertEqual(spval, sk_pval)


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import coloredlogs
import logging
import os
import sys
import dill as pickle

from cyclopts import App, Parameter
import numpy as np
import pandas as pd
from typing import List
from typing_extensions import Annotated

from neurostatx.io.utils import (load_df_in_any_format, assert_input,
                                 assert_output_dir_exist)
from neurostatx.io.viz import flexible_barplot
from neurostatx.clustering.fuzzy import fuzzyCmeans
from neurostatx.utils.preprocessing import merge_dataframes, compute_pca
from neurostatx.clustering.viz import (
    plot_clustering_results,
    plot_dendrogram,
    plot_parallel_plot,
    radar_plot
)
from neurostatx.clustering.metrics import (compute_knee_location,
                                           find_optimal_gap)
from neurostatx.clustering.distance import DistanceMetrics


# Initializing the app.
app = App(default_parameter=Parameter(negative=()))


@app.default()
def FuzzyClustering(
    in_dataset: Annotated[
        List[str],
        Parameter(
            show_default=False,
            group="Essential Files Options",
        ),
    ],
    id_column: Annotated[
        str,
        Parameter(
            show_default=False,
            group="Essential Files Options",
        ),
    ],
    desc_columns: Annotated[
        int,
        Parameter(
            show_default=False,
            group="Essential Files Options",
        ),
    ],
    k: Annotated[
        int,
        Parameter(
            show_default=True,
            group="Clustering Options",
        ),
    ] = 10,
    m: Annotated[
        float,
        Parameter(
            show_default=True,
            group="Clustering Options",
        ),
    ] = 2,
    error: Annotated[
        float,
        Parameter(
            show_default=True,
            group="Clustering Options",
        ),
    ] = 1e-6,
    maxiter: Annotated[
        int,
        Parameter(
            show_default=True,
            group="Clustering Options",
        ),
    ] = 1000,
    init: Annotated[
        str,
        Parameter(
            show_default=True,
            group="Clustering Options",
        ),
    ] = None,
    metric: Annotated[
        DistanceMetrics,
        Parameter(
            show_default=True,
            group="Clustering Options",
        ),
    ] = DistanceMetrics.euclidean,
    pca: Annotated[
        bool,
        Parameter(
            "--pca",
            show_default=True,
            group="Clustering Options",
        ),
    ] = False,
    out_folder: Annotated[
        str,
        Parameter(
            group="Essential Files Options",
        ),
    ] = "./fuzzy_results/",
    processes: Annotated[
        int,
        Parameter(
            show_default=True,
            group="Computational Options",
        ),
    ] = 1,
    parallelplot: Annotated[
        bool,
        Parameter(
            show_default=True,
            group="Visualization Options",
        ),
    ] = False,
    radarplot: Annotated[
        bool,
        Parameter(
            show_default=True,
            group="Visualization Options",
        ),
    ] = True,
    cmap: Annotated[
        str,
        Parameter(
            show_default=True,
            group="Visualization Options",
        ),
    ] = "magma",
    verbose: Annotated[
        bool,
        Parameter(
            "-v",
            "--verbose",
            group="Optional parameters",
        ),
    ] = False,
    save_parameters: Annotated[
        bool,
        Parameter(
            "-s",
            "--save_parameters",
            group="Optional parameters",
        ),
    ] = False,
    overwrite: Annotated[
        bool,
        Parameter(
            "-f",
            "--overwrite",
            group="Optional parameters",
        ),
    ] = False,
):
    """FUZZY CLUSTERING
    ----------------
    FuzzyClustering is a wrapper script for a Fuzzy C-Means
    clustering analysis. By design, the script will compute the analysis for
    k specified cluster (chosen by --k) and returns various
    evaluation metrics and summary barplot/parallel plot.

    EVALUATION METRICS
    ------------------
    The fuzzy partition coefficient (FPC) is a metric defined between 0 and 1
    with 1 representing the better score. It represents how well the data is
    described by the clustering model. Therefore, a higher FPC represents a
    better fitted model. On real-world data, local maxima can also be
    interpreted as one of the optimal solution.

    The Silhouette Coefficient represents an evaluation of cluster's
    definition. The score is bounded (-1 to 1) with 1 as the perfect score and
    -1 as not a good clustering result. A higher Silhouette Coefficient relates
    to a model with better defined clusters (therefore a better model). It
    tends to have higher score with cluster generated from density-
    based methods.

    The Calinski-Harabasz Index (or the Variance Ratio Criterion) can be used
    when no known labels are available. It represents the density and
    separation of clusters. Although it tends to be higher for cluster
    generated from density-based methods. A higher Calinski-Harabasz Index
    relates to better defined clusters.

    Davies-Bouldin Index is reported for all cluster-models. A lower DBI
    relates to a model with better cluster separation. It represents a measure
    of similarity between clusters and is solely based on quantities and
    features of the dataset. It also tends to be generally higher for
    convex clusters and it uses the centroid distance between clusters
    therefore limiting the distance metric to euclidean space.

    Within cluster Sum of Squared error (WSS) represents the average distance
    from each point to their cluster centroid. WSS is combined with the elbow
    method to determine the optimal k number of clusters.

    The GAP statistics is based on the WSS. It relies on computing the
    difference in cluster compactness between the actual data and simulated
    data with a null distribution. The optimal k-number of clusters is
    identified by a maximized GAP statistic (local maxima can also suggest
    possible solutions.).

    PARAMETERS
    ----------
    Details regarding the parameters can be seen below. Regarding the
    --m parameter, it defines the degree of fuzziness of the resulting
    membership matrix. Using --m 1 will returns crisp clusters, whereas
    --m >1 will returned more and more fuzzy clusters. It is also possible
    to pre-initialize the c-partitioned matrix from previous membership matrix.
    If you want to do that, you need to specify a folder containing all
    membership matrices for each k number (meaning that if you want to perform
    clustering up to k=10, you need a membership matrices for each of them.).
    It also must respect this name convention:
    ::

                    [init_folder]
                        |-- cluster_membership_1.npy
                        |-- cluster_membership_2.npy
                        |-- [...]
                        └-- cluster_membership_{k}.npy

    OUTPUT FOLDER STRUCTURE
    -----------------------
    The script creates a default output structure in a destination specified
    by using --out-folder. Output structure is as follows:
    ::

                    [out_folder]
                        |-- CENTROIDS
                        |       |-- clusters_centroids_2.xlsx
                        |       |-- [...]
                        |       └-- clusters_centroids_{k}.xlsx
                        |-- MEMBERSHIP_DF
                        |       |-- clusters_membership_2.xlsx
                        |       |-- [...]
                        |       └-- clusters_membership_{k}.xlsx
                        |-- MEMBERSHIP_MAT (in .npy format)
                        |-- METRICS
                        |       |-- chi.png
                        |       |-- [...]
                        |       └-- wss.png
                        |-- PARALLEL_PLOTS (optional)
                        |       |-- parallel_plot_2clusters.png
                        |       |-- [...]
                        |       |-- parallel_plot_{k}clusters.png
                        |-- PCA (optional)
                        |       |-- transformed_data.xlsx
                        |       |-- variance_explained.xlsx
                        |       └-- pca_model.joblib
                        |-- RADAR_PLOTS (optional)
                        |       |-- radar_plot_2clusters.png
                        |       |-- [...]
                        |       |-- radar_plot_{k}clusters.png
                        |-- validation_indices.xlsx
                        └-- viz_multiple_cluster_nb.png

    REFERENCES
    ----------
    [1] [Scikit-Fuzzy
    Documentation](https://pythonhosted.org/scikit-fuzzy/auto_examples/plot_cmeans.html)

    [2] [Scikit-Learn Documentation - Clustering Performance
    Evaluation](https://scikit-learn.org/stable/modules/clustering.html#clustering-performance-evaluation)

    [3] [Selecting the optimal number of clusters -
    1](https://towardsdatascience.com/cheat-sheet-to-implementing-7-methods-for-selecting-optimal-number-of-clusters-in-python-898241e1d6ad)

    [4] [Selecting the optimal number of clusters -
    2](https://towardsdatascience.com/how-to-determine-the-right-number-of-clusters-with-code-d58de36368b1)

    [5] [Scikit-Fuzzy GitHub
    Repository](https://github.com/scikit-fuzzy/scikit-fuzzy)

    EXAMPLE USAGE
    -------------
    ::

        FuzzyClustering --in-dataset dataset.csv --id-column ID --desc-columns
        1 --k 10 --m 2 --error 1e-6 --maxiter 1000 --init init_folder --metric
        euclidean --pca --out-folder ./fuzzy_results/ --processes 4 --verbose

    Parameters
    ----------
    in_dataset : List[str]
        Input dataset(s) to filter. If multiple files are provided as input,
        will be merged according to the subject id columns.
    id_column : str
        Name of the column containing the subject's ID tag. Required for
        proper handling of IDs and merging multiple datasets.
    desc_columns : int
        Number of descriptive columns at the beginning of the dataset to
        exclude in statistics and descriptive tables.
    k : int, optional
        Maximum k number of cluster to fit a model for. (Script will iterate
        until k is met.)
    m : float, optional
        Exponentiation value to apply on the membership function, will
        determined the degree of fuzziness of the membership matrix
    error : float, optional
        Error threshold for convergence stopping criterion.
    maxiter : int, optional
        Maximum number of iterations to perform.
    init : str, optional
        Initial fuzzy c-partitioned matrix
    metric : DistanceMetrics, optional
        Metric to use to compute distance between original points and clusters
        centroids.
    pca : bool, optional
        If set, will perform PCA decomposition to 2 components before
        clustering.
    out_folder : str, optional
        Path of the folder in which the results will be written. If not
        specified, current folder and default name will be used.
    processes : int, optional
        Number of processes to launch in parallel.
    parallelplot : bool, optional
        If true, will output parallel plot for each cluster solution. Default
        is False.
    radarplot : bool, optional
        If true, will output radar plot for each cluster solution. Default is
        True.
    cmap : str, optional
        Colormap to use for plotting. Default is "magma". See
        https://matplotlib.org/stable/tutorials/colors/colormaps.html.
    verbose : bool, optional
        If true, produce verbose output.
    save_parameters : bool, optional
        If true, will save input parameters to .txt file.
    overwrite : bool, optional
        If true, force overwriting of existing output files.
    """

    if verbose:
        logging.getLogger().setLevel(logging.INFO)
        coloredlogs.install(level=logging.INFO)

    assert_input(in_dataset)
    assert_output_dir_exist(overwrite, out_folder, create_dir=True)

    if save_parameters:
        parameters = list(locals().items())
        with open(f"{out_folder}/parameters.txt", "w+") as f:
            for param in parameters:
                f.writelines(str(param))

    # Creating substructures for output folder.
    os.mkdir(f"{out_folder}/METRICS/")

    # Loading dataframe.
    logging.info("Loading dataset(s)...")
    if len(in_dataset) > 1:
        if id_column is None:
            sys.exit(
                "Column name for index matching is required when inputting "
                "multiple dataframe."
            )
        dict_df = {i: load_df_in_any_format(i) for i in in_dataset}
        raw_df = merge_dataframes(dict_df, id_column)
    else:
        raw_df = load_df_in_any_format(in_dataset[0])
    descriptive_columns = [n for n in range(0, desc_columns)]

    # Creating the array.
    desc_data = raw_df[raw_df.columns[descriptive_columns]]
    df_for_clust = raw_df.drop(
        raw_df.columns[descriptive_columns], axis=1, inplace=False
    ).astype("float")
    X = df_for_clust.values

    # Decomposing into 2 components if asked.
    if pca:
        logging.info("Applying PCA dimensionality reduction.")
        X, model, variance, components, chi, kmo = compute_pca(X, 3)
        logging.info(
            "Bartlett's test of sphericity returned a p-value of {} and "
            "Keiser-Meyer-Olkin (KMO)"
            " test returned a value of {}.".format(chi, kmo)
        )

        # Exporting variance explained data.
        os.mkdir(f"{out_folder}/PCA/")
        var_exp = pd.DataFrame(variance, columns=["Variance Explained"])
        var_exp.to_excel(
            f"{out_folder}/PCA/variance_explained.xlsx", index=True,
            header=True
        )
        components_df = pd.DataFrame(components, columns=df_for_clust.columns)
        components_df.to_excel(f"{out_folder}/PCA/components.xlsx", index=True,
                               header=True)
        out = pd.DataFrame(X, columns=["Component #1", "Component #2",
                                       "Component #3"])
        out = pd.concat([desc_data, out], axis=1)
        out.to_excel(f"{out_folder}/PCA/transformed_data.xlsx", index=True,
                     header=True)

        flexible_barplot(
            components_df.T,
            3,
            f"{out_folder}/PCA/barplot_loadings.png",
            cmap=cmap,
            title="Loadings values for the three components.",
            ylabel="Loading values")

        # Exporting model in .joblib format.
        with open(f"{out_folder}/PCA/pca_model.pkl", "wb") as f:
            pickle.dump(model, f)

    # Plotting the dendrogram.
    logging.info("Generating dendrogram.")
    sys.setrecursionlimit(50000)
    plot_dendrogram(X, f"{out_folder}/METRICS/dendrogram.png")

    # Load initialisation matrix if any.
    if init is not None:
        init_mat = [
            np.load(f"{init}/clusters_membership_{i}.npy")
            for i in range(2, k + 1)
        ]
    else:
        init_mat = None

    # Computing a range of C-means clustering method.
    logging.info("Computing FCM from k=2 to k={}".format(k))
    cntr, u, wss, fpcs, ss, chi, dbi, gap, sk = fuzzyCmeans(
        X,
        max_cluster=k,
        m=m,
        error=error,
        maxiter=maxiter,
        init=init_mat,
        metric=metric,
        output=out_folder,
        processes=processes,
        verbose=verbose,
    )

    # Compute knee location on Silhouette Score.
    logging.info("Plotting validation indicators and outputting final "
                 "matrices.")
    elbow_wss = compute_knee_location(wss)

    # Creating a dataframe to export statistics.
    stats = pd.DataFrame(
        data={
            "FPC": fpcs,
            "Silhouette Score": ss,
            "CHI": chi,
            "DBI": dbi,
            "WSS": wss,
            "GAP": gap,
        },
        index=[f"{i}-Cluster Model" for i in range(2, len(ss) + 2)],
    )
    stats.to_excel(f"{out_folder}/validation_indices.xlsx", header=True,
                   index=True)

    # Plotting results for each indicators.
    plot_clustering_results(
        wss,
        title="Within Cluster Sum of Square Error (WSS)",
        metric="WSS",
        output=f"{out_folder}/METRICS/wss.png",
        annotation=f"Elbow threshold (Optimal cluster nb): {elbow_wss}",
    )
    fpcs_index = fpcs.index(max(fpcs))
    plot_clustering_results(
        fpcs,
        title="Fuzzy Partition Coefficient (FPC)",
        metric="FPC",
        output=f"{out_folder}/METRICS/fpc.png",
        annotation=f"Optimal Number of Cluster: {fpcs_index+2}",
    )
    ss_index = ss.index(max(ss))
    plot_clustering_results(
        ss,
        title="Silhouette Score Coefficient (SS)",
        metric="SS",
        output=f"{out_folder}/METRICS/ss.png",
        annotation=f"Optimal Number of Clusters: {ss_index+2}",
    )
    chi_index = chi.index(max(chi))
    plot_clustering_results(
        chi,
        title="Calinski-Harabasz Index (CHI)",
        metric="CHI",
        output=f"{out_folder}/METRICS/chi.png",
        annotation=f"Optimal Number of Clusters: {chi_index+2}",
    )
    dbi_index = dbi.index(min(dbi))
    plot_clustering_results(
        dbi,
        title="Davies-Bouldin Index (DBI)",
        metric="DBI",
        output=f"{out_folder}/METRICS/dbi.png",
        annotation=f"Optimal Number of Clusters: {dbi_index+2}",
    )
    gap_index = find_optimal_gap(gap, sk)
    plot_clustering_results(
        gap,
        title="GAP Statistics.",
        metric="GAP",
        output=f"{out_folder}/METRICS/gap.png",
        errorbar=sk,
        annotation=f"Optimal Number of Clusters: {gap_index+2}",
    )

    # Exporting plots and graphs for each cluster solution.
    os.mkdir(f"{out_folder}/MEMBERSHIP_MAT/")
    os.mkdir(f"{out_folder}/MEMBERSHIP_DF/")
    os.mkdir(f"{out_folder}/PARALLEL_PLOTS/")
    os.mkdir(f"{out_folder}/RADAR_PLOTS/")
    os.mkdir(f"{out_folder}/CENTROIDS/")

    # Iterating and saving every elements.
    for i in range(len(u)):
        membership = np.argmax(u[i], axis=0)
        if parallelplot:
            plot_parallel_plot(
                df_for_clust,
                membership,
                mean_values=True,
                output=f"{out_folder}/PARALLEL_PLOTS/parallel_plot_{i+2}"
                       "clusters.png",
                cmap=cmap,
                title=f"Parallel Coordinates plot for {i+2} clusters solution."
            )
        if radarplot:
            radar_plot(
                df_for_clust,
                membership,
                title=f"Radar plot of {i+2} clusters solution.",
                frame='circle',
                cmap=cmap,
                output=f"{out_folder}/RADAR_PLOTS/radar_plot_{i+2}clusters.png"
            )

        # Converting membership arrays to df.
        member = pd.DataFrame(
            u[i].T,
            index=None,
            columns=[f"Cluster #{n+1}" for n in range(u[i].shape[0])],
        )
        centroids = pd.DataFrame(
            cntr[i],
            index=[f"Cluster #{n+1}" for n in range(u[i].shape[0])],
            columns=[f"v{i}" for i in range(X.shape[1])],
        )

        # Appending subject ids and descriptive columns.
        member_out = pd.concat([raw_df, member], axis=1)
        member_out.to_excel(
            f"{out_folder}/MEMBERSHIP_DF/clusters_membership_{i+2}.xlsx",
            header=True,
            index=False,
        )
        centroids.to_excel(
            f"{out_folder}/CENTROIDS/clusters_centroids_{i+2}.xlsx",
            header=True,
            index=True,
        )

        # Saving original matrix.
        np.save(f"{out_folder}/MEMBERSHIP_MAT/clusters_membership_{i+2}.npy",
                u[i])


if __name__ == "__main__":
    app()



from enum import Enum
import networkx as nx
import numpy as np
import matplotlib.pyplot as plt
from matplotlib.cm import ScalarMappable
import matplotlib
import seaborn as sns
from strenum import StrEnum

from neurostatx.network.utils import (filter_node_centroids,
                                      filter_node_subjects)


class NetworkLayout(StrEnum, Enum):
    KamadaKawai = ("kamada_kawai_layout",)
    Spectral = ("spectral_layout",)
    Spring = "spring_layout"


def compute_layout(G,
                   layout=nx.spring_layout,
                   weight="weight",
                   seed=1234):
    """Function to compute the optimal layout for a graph network.

    Args:
        G (nx.Graph):               Graph Network to draw.
        layout (nx.layout, optional):  Layout algorithm to use.
                                    Defaults to nx.spring_layout.
        weight (str, optional):     Edges weight to use while computing the
                                    layout. Defaults to "weight".
        seed (int, optional):       Random seed. Defaults to 1234.

    Returns:
        pos:                        Layout positions that can be reused for
                                    replotting of the same network.
    """

    return layout(G, weight=weight, seed=seed)


def set_nodes_position(G, pos):
    """Function to set the nodes' position for a graph network.

    Position array has to be converted into python float in order for
    Gephi file format export.

    Args:
        G (nx.Graph):       Graph Network to draw.
        pos (dict):         Dictionary with nodes' positions.

    Returns:
        G:                  Network with nodes' positions set.
    """
    positions = {k: list(map(float, pos[k])) for k in pos}

    return nx.set_node_attributes(G, positions, "pos")


def visualize_network(
    G,
    output,
    weight="weight",
    centroids_labelling=True,
    subjects_labelling=False,
    centroid_node_shape=500,
    centroid_alpha=1,
    centroid_node_color="white",
    centroid_edge_color="black",
    subject_node_shape=5,
    subject_alpha=0.3,
    subject_node_color="black",
    subject_edge_color=None,
    colormap="plasma",
    title="Graph Network",
    legend_title="Membership values",
):
    """
    Function to visualize a weighted undirected graph network. Based on the
    concept from:
    Ariza-Jiménez, L., Villa, L. F., & Quintero, O. L. (2019). Memberships
        Networks for High-Dimensional Fuzzy Clustering Visualization.,
        Applied Computer Sciences in Engineering (Vol. 1052, pp. 263–273).
        Springer International Publishing.
        https://doi.org/10.1007/978-3-030-31019-6_23

    Args:
        G (NetworkX.Graph()):                       Graph Network to draw.
        output (str):                               Filename and path for
                                                    the output png image.
        weight (str, optional):                     Edge attribute to use as
                                                    weight. Defaults to
                                                    'weight'.
        centroids_labelling (bool, optional):       Label centroid nodes.
                                                    Defaults to True.
        subjects_labelling (bool, optional):        Label subject nodes.
                                                    Defaults to False.
        centroid_node_shape (int, optional):        Centroid's nodes shape.
                                                    Defaults to 500.
        centroid_alpha (int, optional):             Centroid's nodes alpha.
                                                    Defaults to 1.
        centroid_node_color (str, optional):        Centroid's nodes color.
                                                    Defaults to 'white'.
        centroid_edge_color (str, optional):        Centroid's nodes edge
                                                    color. Defaults to 'black'.
        subject_node_shape (int, optional):         Subject's nodes shape.
                                                    Defaults to 5.
        subject_alpha (float, optional):            Subject's nodes alpha
                                                    value.Defaults to 0.3.
        subject_node_color (str, optional):         Subject's nodes color.
                                                    Defaults to 'black'.
        subject_edge_color (_type_, optional):      Subject's nodes edge color.
                                                    Defaults to None.
        colormap (str, optional):                   Colormap to use to draw
                                                    edges' weights. Defaults
                                                    to 'plasma'.
        title (str, optional):                      Graph title. Defaults to
                                                    'Graph Network'.
        legend_title (str, optional):               Legend title. Defaults to
                                                    'Membership values'.
    """

    # Fetching nodes position.
    pos = nx.get_node_attributes(G, "pos")

    # Fetching edges widths.
    widths = nx.get_edge_attributes(G, weight)

    # Sorting which nodes to label.
    labels = {}
    if centroids_labelling:
        for node in G.nodes():
            if "c" in node:
                labels[node] = node
    elif centroids_labelling and subjects_labelling:
        for node in G.nodes():
            labels[node] = node
    else:
        for node in G.nodes():
            labels[node] = ""

    # Setting z-order of nodes.
    cntr_node = nx.subgraph_view(G, filter_node=filter_node_centroids)
    sub_node = nx.subgraph_view(G, filter_node=filter_node_subjects)

    # Centroids customization lists.
    cntr_shape = np.array([centroid_node_shape] * len(cntr_node.nodes()))
    cntr_alpha = np.array([centroid_alpha] * len(cntr_node.nodes()))

    # Subjects customization lists.
    sub_shape = np.array([subject_node_shape] * len(sub_node.nodes()))
    # sub_alpha = np.array([subject_alpha] * len(sub_node.nodes()))

    # Plotting the graph.
    fig = plt.figure(figsize=(12, 8))
    ax = fig.add_subplot()

    nodes1 = nx.draw_networkx_nodes(
        G,
        pos,
        nodelist=sub_node.nodes(),
        node_size=sub_shape,
        node_color=subject_node_color,
        alpha=subject_alpha,
        ax=ax,
    )
    nodes2 = nx.draw_networkx_nodes(
        G,
        pos,
        nodelist=cntr_node.nodes(),
        node_size=cntr_shape,
        node_color=centroid_node_color,
        alpha=cntr_alpha,
        ax=ax,
    )

    # Drawing edges.
    nx.draw_networkx_edges(
        G,
        pos,
        edgelist=widths.keys(),
        width=(list(widths.values()) * 10),
        edge_color=list(widths.values()),
        edge_cmap=getattr(plt.cm, colormap),
        alpha=list(widths.values()),
        ax=ax,
    )

    # Setting z-order.
    nodes1.set_zorder(2)
    if subject_edge_color is not None:
        nodes1.set_edgecolor(subject_edge_color)
    nodes2.set_zorder(3)
    nodes2.set_edgecolor(centroid_edge_color)

    # Plotting labels if set.
    nx.draw_networkx_labels(G, pos, labels=labels, font_color="black", ax=ax)

    # Adding colorbar, titles, etc.
    cmappable = ScalarMappable(matplotlib.colors.Normalize(0, 1),
                               getattr(plt.cm, colormap))
    cbar = plt.colorbar(cmappable, ax=ax, location="right", shrink=0.5)

    plt.box(False)
    ax.set_title(title)
    cbar.ax.set_title(legend_title)

    plt.tight_layout()
    plt.savefig(output)
    plt.close()


def membership_distribution(mat, output="./membership_distribution.png"):
    """
    Function returning a plot of the distribution of the maximum membership
    values for each subject and the delta plot of the difference with the
    second highest membership from the fuzzy c-partitioned matrix.

    Args:
        mat (Array):                    Membership fuzzy c-partitioned matrix.
        output (str, optional):         Output filename and path. Defaults to
                                        './membership_distribution.png'.
    """

    # Fetching 1st highest membership value.
    high1st = np.max(mat, axis=1)

    # Fetching 2nd highest membership value.
    high2nd = np.partition(mat, -2, axis=1)[:, -2]

    # Calculating delta.
    delta = high1st - high2nd

    # Plotting the distribution.
    with plt.rc_context(
        {"font.family": "Sans Serif",
         "font.size": 12, "font.weight": "normal", "axes.titleweight": "bold"}
    ):

        fig, ax = plt.subplots(2, 2, figsize=(16, 16))

        # Plotting 1st highest membership value.
        sns.histplot(data=high1st, stat="density", bins=50, kde=True,
                     ax=ax[0, 0], color="lightgray")
        ax[0, 0].set_xlabel(
            "1st highest membership value distribution amongst all subjects."
        )
        ax[0, 0].axvline(x=np.median(high1st), ymin=0, ymax=1, color="gray",
                         linestyle="--")

        # Plotting 2nd highest memberhsip value.
        sns.histplot(data=high2nd, stat="density", bins=50, kde=True,
                     ax=ax[0, 1], color="lightgray")
        ax[0, 1].set_xlabel(
            "2nd highest membership value distribution amongst all subjects."
        )
        ax[0, 1].axvline(x=np.median(high2nd), ymin=0, ymax=1, color="gray",
                         linestyle="--")

        # Plotting delta.
        sns.histplot(data=delta, stat="density", bins=50, kde=True,
                     ax=ax[1, 0], color="lightgray")
        ax[1, 0].set_xlabel(
            "Delta (1st highest value - 2nd highest value) distribution "
            "amongst all subjects."
        )
        ax[1, 0].axvline(x=np.median(delta), ymin=0, ymax=1, color="gray",
                         linestyle="--")

        ax[1, 1].remove()
        plt.tight_layout()
        plt.savefig(f"{output}")
        plt.close()


def creating_node_colormap(percentile_dict):
    """
    Function to create a colormap for a set of nodes based on a percentile
    dictionary.

    Args:
        percentile_dict (dict):     Percentile dictionary with keys identifying
                                    clusters.

    Return:
        cmap:                       List containing all colors.
    """

    cmap = plt.cm.tab10(np.linspace(0, 1, 10))

    # Assuming it is impossible for node to have 2 main clusters.
    percentile_array = np.max(np.array(list(percentile_dict.values())), axis=0)

    nodes_cmap = []
    for i in percentile_array:
        if i == 0:
            nodes_cmap.append("black")
        else:
            nodes_cmap.append(cmap[i])

    return nodes_cmap


def create_cmap_from_list(array):
    """
    Function to create a cmap from a list of values (corresponding to a
    condition for each subject.) For example, subject with value = 0 will be
    mapped black, whereas subject = >1 will be mapped to different colors.

    Args:
        array (Array):      Array of integers corresponding to condiditions
                            (can be binary also.).

    Return:
        cmap:               List containing all colors.
    """

    if array.dtype == "float64":
        cmap = plt.cm.plasma(np.linspace(0, 1, len(array)))

        ranks = array.argsort().argsort()

        nodes_cmap = []
        for i in ranks:
            nodes_cmap.append(cmap[i])

    else:
        cmap = plt.cm.tab10(np.linspace(0, 1, 10))

        nodes_cmap = []
        for i in array:
            if i == 0:
                nodes_cmap.append("darkgrey")
            else:
                nodes_cmap.append(cmap[i])

    return nodes_cmap



# !/usr/bin/env python3
# -*- coding: utf-8 -*-

import os
import tempfile

from neurostatx.io.download import get_home, download_data, save_files_dict


download_data(save_files_dict(), keys=["data.zip"])
tmp_dir = tempfile.TemporaryDirectory()


def test_help(script_runner):
    ret = script_runner.run(["ComputeGraphNetwork", "-h"])

    assert ret.success


def test_compute_graph_network(script_runner):
    os.chdir(os.path.expanduser(tmp_dir.name))
    in_matrix = os.path.join(get_home(), "data/clusters_membership_3.xlsx")
    out_folder = os.path.join(get_home(), "data/graph_network/")

    ret = script_runner.run([
        "ComputeGraphNetwork",
        "--in-dataset", in_matrix,
        "--out-folder", out_folder,
        "--id-column", "subjectkey",
        "--plot-distribution",
        "--import-data",
        "--desc-columns", 1, "-f"]
    )

    assert ret.success



# -*- coding: utf-8 -*-

from collections import OrderedDict
from enum import Enum
import logging
import multiprocessing
import random

from functools import partial
import networkx as nx
import numpy as np
from p_tqdm import p_map


def get_metrics_ops():
    """
    Get a dictionary of all functions related to graph network metrics.

    Returns:
        OrderedDict:    Functions dictonary.
    """
    return OrderedDict(
        [
            ("eigencentrality", eigencentrality),
            ("closenesscentrality", closenesscentrality),
            ("betweennesscentrality", betweennesscentrality),
            ("informationcentrality", informationcentrality),
            ("currentflowbc", currentflowbc),
            ("loadcentrality", loadcentrality),
            ("harmoniccentrality", harmoniccentrality),
            ("eccentricity", eccentricity),
            ("clustering", clustering),
            ("constraint", constraint),
            ("effectivesize", effectivesize),
            ("closenessvitality", closenessvitality),
            ("degree", degree),
        ]
    )


def get_metrics_docs(ops: dict):
    """
    Function to fetch from a dictionary all functions and join all
    documentations.

    Args:
        ops (dict):         Dictionary of functions.
    """
    full_doc = []
    for func in ops.values():
        full_doc.append(func.__doc__)
    return "\n".join(full_doc)


def eigencentrality(graph, weight=None):
    """
    eigencentrality: GRAPH WEIGHT\n
        Will return a dictionary of the eigenvector centrality for all nodes.
    """
    return nx.eigenvector_centrality(graph, max_iter=1000, weight=weight)


def closenesscentrality(graph, weight=None):
    """
    closenesscentrality: GRAPH WEIGHT\n
        Will return a dictionary of the closeness centrality for all nodes.
    """
    return nx.closeness_centrality(graph, distance=weight, wf_improved=True)


def betweennesscentrality(graph, weight=None):
    """
    betweennesscentrality: GRAPH WEIGHT\n
        Will return a dictionary of the betweenness centrality for all nodes.
    """
    return nx.betweenness_centrality(graph, weight=weight)


def informationcentrality(graph, weight=None):
    """
    informationcentrality: GRAPH WEIGHT\n
        Will return a dictionary of the information centrality for all nodes.
    """
    return nx.information_centrality(graph, weight=weight, solver="full")


def currentflowbc(graph, weight=None):
    """
    currentflowbc: GRAPH WEIGHT\n
        Will return a dictionary of the current flow betweenness centrality
        for all nodes.
    """
    return nx.current_flow_betweenness_centrality(graph, weight=weight,
                                                  solver="full")


def loadcentrality(graph, weight=None):
    """
    loadcentrality: GRAPH WEIGHT\n
        Will return a dictionary of the load centrality for all nodes.
    """
    return nx.load_centrality(graph, weight=weight)


def harmoniccentrality(graph, weight=None):
    """
    harmoniccentrality: GRAPH WEIGHT\n
        Will return a dictionary of the harmonic centrality for all nodes.
    """
    return nx.harmonic_centrality(graph, distance=weight)


def eccentricity(graph, weight=None):
    """
    eccentricity: GRAPH WEIGHT\n
        Will return a dictionary of the eccentricity for all nodes.
    """
    return nx.eccentricity(graph, weight=weight)


def clustering(graph, weight=None):
    """
    clustering: GRAPH WEIGHT\n
        Will return a dictionary of the clustering coefficient for all nodes.
    """
    return nx.clustering(graph, weight=weight)


def constraint(graph, nodes, weight=None):
    """
    constraint: GRAPH NODES WEIGHT\n
        Will return a dictionary of the constraint for all specified nodes.
    """
    return nx.constraint(graph, nodes=nodes.split(), weight=weight)


def effectivesize(graph, nodes, weight=None):
    """
    effectivesize: GRAPH NODES WEIGHT\n
        Will return a dictionary of the effective size for all specified nodes.
    """
    return nx.effective_size(graph, nodes=nodes.split(), weight=weight)


def closenessvitality(graph, nodes, weight=None):
    """
    closenessvitality: GRAPH NODE WEIGHT\n
        Will return a dictionary of the closeness vitality for a single node.
    """
    return nx.closeness_vitality(graph, node=nodes, weight=weight)


def degree(graph, weight=None):
    """
    degree: GRAPH WEIGHT\n
        Will return the degree of the specified node.
    """
    return graph.degree(weight=weight)


class PathLengthsMethods(str, Enum):
    Dijkstra = ("dijkstra",)
    BellmanFord = ("bellman-ford",)
    FloydWarshall = ("floyd-warshall",)
    FloydWarshallNumpy = "floyd-warshall-numpy"


def weightedpath(
    graph,
    df,
    label_name,
    cohort=None,
    iterations=1000,
    weight=None,
    method="dijkstra",
    distribution=None,
    processes=1,
    verbose=False,
):
    """
    Function to compute the average weighted shortest path length for a group
    of nodes. The function will also compute the p-value between the group of
    nodes and the randomly generated null distribution.

    Args:
        graph (Networkx.graph):                 Networkx graph object.
        df (pandas.DataFrame):                  Dataframe containing the nodes
        label_name (str):                       Name of the column containing
                                                the group label.
        iterations (int, optional):             Number of iterations to run.
                                                Defaults to 1000.
        weight (str, optional):                 Edge attributes to use as
                                                weight. Defaults to None.
        method (str, optional):                 Method to use for path
                                                computation. Defaults to
                                                "dijkstra".
        distribution (pd.DataFrame, optional):  Pre-computed distribution.
                                                Defaults to None.
        processes (int, optional):              Number of processes to use.
                                                Defaults to 1.
        verbose (bool, optional):               Verbose flag. Defaults to
                                                False.

    Returns:
        avg_path_length (float):                Average path length.
        dist (list):                            Null distribution.
        pvalue (float):                         P-value.
    """
    # Cohort selection.
    if cohort is not None:
        subset = df.loc[df['cohort'] == cohort]
    else:
        subset = df.copy()

    # Setting lists.
    group_exclude = subset.loc[subset[label_name] == 0]
    nodes_exclude = group_exclude.index.to_list()
    nodes_include = [node for node in list(subset.index) if node
                     not in nodes_exclude]
    centroid_nodes = [node for node in list(graph) if 'c' in node]
    nodes_include = nodes_include + centroid_nodes

    logging.info("Computing weighted path for the set of nodes.")
    sub_G = nx.induced_subgraph(graph, nodes_include)
    avg_path_length = nx.average_shortest_path_length(
        sub_G, weight=weight, method=method
    )

    # Fetching all possible nodes.
    nodes_list = df.index.to_list()

    # Setting partial function to pass common arguments between iterations.
    if distribution is None:
        generate_null_dist = partial(
            _weightedpath,
            graph,
            nodes_list=nodes_list,
            sample_size=len(nodes_include),
            weight=weight,
            method=method,
        )

        # Opening multiprocessing pool.
        logging.info("Computing null distribution.")
        multiprocessing.set_start_method("spawn", force=True)
        pool = multiprocessing.Pool(processes=processes)

        # Initiating processing.
        if verbose:
            dist = p_map(generate_null_dist, range(0, iterations))
        else:
            dist = pool.map(generate_null_dist, range(0, iterations))
        pool.close()
        pool.join()
    else:
        dist = distribution[label_name].values

    # Compute p-value.
    pvalue = (
        (np.sum(np.array(dist) >= avg_path_length) + 1) / (iterations + 1)
    )

    return avg_path_length, dist, pvalue


def _weightedpath(
    graph, n_iter, nodes_list, sample_size, weight=None, method="dijkstra"
):
    """
    Core worker of weightedpath() function.
    """

    # Filtering nodes.
    random.shuffle(nodes_list)
    random_nodes = random.sample(nodes_list, sample_size)
    nodes_exclude = list(set(nodes_list) - set(random_nodes))

    # Copying graph.
    # orig = graph.copy()

    # Filtering original graph nodes to remove the ones that were not randomly
    # selected (easier to keep cluster centroids nodes that way.).
    # sub_G = nx.induced_subgraph(orig,
    #                            list(set(list(orig)) - set(nodes_exclude)))
    sub_G = graph.subgraph(list(set(graph) - set(nodes_exclude)))

    # Computing weighted path length.
    weighted_avg_path_length = nx.average_shortest_path_length(
        sub_G, weight=weight, method=method
    )

    return weighted_avg_path_length
